# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
from typing import Optional

import torch
import torch.nn.functional as F

# Activations


def max_min(input: torch.Tensor, dim: Optional[int] = None) -> torch.Tensor:
    r"""
    Applies max-min activation on the given tensor.

    If ``input`` is a tensor of shape :math:`(N, C)` and ``dim`` is
    ``None``, the output can be described as:

    .. math::
        \text{out}(N_i, C_{2j}) = \max(\text{input}(N_i, C_j), 0)\\
        \text{out}(N_i, C_{2j + 1}) = \max(-\text{input}(N_i, C_j), 0)

    where :math:`N` is the batch size and :math:`C` is the size of the
    tensor.

    Args:
        input:  A tensor of arbitrary shape.
        dim: The dimension to apply max-min. If None, will apply to the
            0th dimension if the shape of input is :math:`(C)` or to the
            first if its :math:`(N, C, *)`.

    Returns:
        A tensor of shape :math:`(2C)` or :math:`(N, 2C, *)` depending
        on the shape of the input.

    Note:
        M. Blot, M. Cord, et N. Thome, « Max-min convolutional neural networks
        for image classification », in 2016 IEEE International Conference on Image
        Processing (ICIP), Phoenix, AZ, USA, 2016, p. 3678‑3682.
    """
    if dim is None:
        if len(input.shape) == 1:
            dim = 0
        else:
            dim = 1
    return torch.cat((F.relu(input), F.relu(-input)), dim=dim)


def group_sort(
    input: torch.Tensor, group_size: Optional[int] = None, dim: int = 1
) -> torch.Tensor:
    r"""
    Applies GroupSort activation on the given tensor.

    See Also:
        :py:func:`group_sort_2`
        :py:func:`full_sort`
    """

    if group_size is None or group_size > input.shape[dim]:
        group_size = input.shape[dim]

    if input.shape[dim] % group_size != 0:
        raise ValueError("The input size must be a multiple of the group size.")

    new_shape = (
        input.shape[:dim]
        + (input.shape[dim] // group_size, group_size)
        + input.shape[dim + 1 :]
    )
    if group_size == 2:
        resh_input = input.view(new_shape)
        a, b = (
            torch.min(resh_input, dim + 1, keepdim=True)[0],
            torch.max(resh_input, dim + 1, keepdim=True)[0],
        )
        return torch.cat([a, b], dim=dim + 1).view(input.shape)
    fv = input.reshape(new_shape)

    return torch.sort(fv, dim=dim + 1)[0].reshape(input.shape)


def group_sort_2(input: torch.Tensor) -> torch.Tensor:
    r"""
    Applies GroupSort-2 activation on the given tensor. This function is equivalent
    to ``group_sort(input, 2)``.

    See Also:
        :py:func:`group_sort`
    """
    return group_sort(input, 2)


def full_sort(input: torch.Tensor) -> torch.Tensor:
    r"""
    Applies FullSort activation on the given tensor. This function is equivalent
    to ``group_sort(input, None)``.

    See Also:
        :py:func:`group_sort`
    """
    return group_sort(input, None)


def lipschitz_prelu(
    input: torch.Tensor, weight: torch.Tensor, k_coef_lip: float = 1.0
) -> torch.Tensor:
    r"""
    Applies k-Lipschitz version of PReLU by clamping the weights

    .. math::
        \text{LPReLU}(x) =
        \begin{cases}
        x, & \text{ if } x \geq 0 \\
        \min(\max(a, -k), k) * x, & \text{ otherwise }
        \end{cases}

    """
    return F.prelu(input, torch.clamp(weight, -k_coef_lip, +k_coef_lip))


# Losses
def apply_reduction(val: torch.Tensor, reduction: str) -> torch.Tensor:
    if reduction == "auto":
        reduction = "mean"
    red = getattr(torch, reduction, None)
    if red is None:
        return val
    return red(val)


def kr_loss(input: torch.Tensor, target: torch.Tensor, multi_gpu=False) -> torch.Tensor:
    r"""
    Loss to estimate the Wasserstein-1 distance using Kantorovich-Rubinstein duality,
    as per

    .. math::
        \mathcal{W}(\mu, \nu) = \sup\limits_{f\in{}Lip_1(\Omega)}
            \underset{\mathbf{x}\sim{}\mu}{\mathbb{E}}[f(\mathbf{x})]
            - \underset{\mathbf{x}\sim{}\nu}{\mathbb{E}}[f(\mathbf{x})]

    where :math:`\mu` and :math:`\nu` are the distributions corresponding to the
    two possible labels as specific by their sign.

    `target` accepts label values in (0, 1), (-1, 1), or pre-processed with the
    `deel.torchlip.functional.process_labels_for_multi_gpu()` function.

    Using a multi-GPU/TPU strategy requires to set `multi_gpu` to True and to
    pre-process the labels `target` with the
    `deel.torchlip.functional.process_labels_for_multi_gpu()` function.

    Args:
        input: Tensor of arbitrary shape.
        target: Tensor of the same shape as input.
        multi_gpu (bool): set to True when running on multi-GPU/TPU

    Returns:
        The Wasserstein-1 loss between ``input`` and ``target``.
    """
    if multi_gpu:
        return kr_loss_multi_gpu(input, target)
    else:
        return kr_loss_standard(input, target)


def kr_loss_standard(input: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
    r"""
    Loss to estimate the Wasserstein-1 distance using Kantorovich-Rubinstein duality,
    as per

    .. math::
        \mathcal{W}(\mu, \nu) = \sup\limits_{f\in{}Lip_1(\Omega)}
            \underset{\mathbf{x}\sim{}\mu}{\mathbb{E}}[f(\mathbf{x})]
            - \underset{\mathbf{x}\sim{}\nu}{\mathbb{E}}[f(\mathbf{x})]

    where :math:`\mu` and :math:`\nu` are the distributions corresponding to the
    two possible labels as specific by their sign.

    `target` accepts label values in (0, 1), (-1, 1)

    Args:
        input: Tensor of arbitrary shape.
        target: Tensor of the same shape as input.

    Returns:
        The Wasserstein-1 loss between ``input`` and ``target``.
    """

    target = target.view(input.shape)
    pos_target = (target > 0).to(input.dtype)
    mean_pos = torch.mean(pos_target, dim=0)
    # pos factor = batch_size/number of positive samples
    pos_factor = torch.nan_to_num(1.0 / mean_pos)
    # neg factor = batch_size/number of negative samples
    neg_factor = -torch.nan_to_num(1.0 / (1.0 - mean_pos))

    weighted_input = torch.where(target > 0, pos_factor, neg_factor) * input
    # Since element-wise KR terms are averaged by loss reduction later on, it is needed
    # to multiply by batch_size here.
    # In binary case (`y_true` of shape (batch_size, 1)), `tf.reduce_mean(axis=-1)`
    # behaves like `tf.squeeze()` to return element-wise loss of shape (batch_size, ).
    return torch.mean(weighted_input, dim=-1)


def kr_loss_multi_gpu(input: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
    r"""Returns the element-wise KR loss when computing with a multi-GPU/TPU strategy.

    `target` and `input` can be either of shape (batch_size, 1) or
    (batch_size, # classes).

    When using this loss function, the labels `target` must be pre-processed with the
    `process_labels_for_multi_gpu()` function.

    Args:
        input: Tensor of arbitrary shape.
        target: pre-processed Tensor of the same shape as input.

    Returns:
        The Wasserstein-1 loss between ``input`` and ``target``.
    """
    target = target.view(input.shape).to(input.dtype)
    # Since the information of batch size was included in `target` by
    # `process_labels_for_multi_gpu()`, there is no need here to multiply by batch size.
    # In binary case (`target` of shape (batch_size, 1)), `torch.mean(dim=-1)`
    # behaves like `torch.squeeze()` to return element-wise loss of shape (batch_size,)
    return torch.mean(input * target, dim=-1)


def neg_kr_loss(
    input: torch.Tensor,
    target: torch.Tensor,
    multi_gpu=False,
) -> torch.Tensor:
    """
    Loss to estimate the negative wasserstein-1 distance using Kantorovich-Rubinstein
    duality.

    Args:
        input: Tensor of arbitrary shape.
        target: Tensor of the same shape as input.
        multi_gpu (bool): set to True when running on multi-GPU/TPU

    Returns:
        The negative Wasserstein-1 loss between ``input`` and ``target``.

    See Also:
        :py:func:`kr_loss`
    """
    return -kr_loss(input, target, multi_gpu=multi_gpu)


def hinge_margin_loss(
    input: torch.Tensor,
    target: torch.Tensor,
    min_margin: float = 1,
) -> torch.Tensor:
    r"""
    Compute the hinge margin loss as per

    .. math::
        \underset{\mathbf{x}}{\mathbb{E}}
        [\max(0, 1 - \mathbf{y} f(\mathbf{x}))]

    Args:
        input: Tensor of arbitrary shape.
        target: Tensor of the same shape as input containing
            target labels (-1 and +1).
        min_margin: The minimal margin to enforce.

    Returns:
        The hinge margin loss.
    """
    target = target.view(input.shape)
    sign_target = torch.where(target > 0, 1.0, -1.0).to(input.dtype)
    hinge = F.relu(min_margin / 2.0 - sign_target * input)
    return torch.mean(hinge, dim=-1)


def hkr_loss(
    input: torch.Tensor,
    target: torch.Tensor,
    alpha: float,
    min_margin: float = 1.0,
    multi_gpu=False,
) -> torch.Tensor:
    """
    Loss to estimate the wasserstein-1 distance with a hinge regularization using
    Kantorovich-Rubinstein duality.

    Args:
        input: Tensor of arbitrary shape.
        target: Tensor of the same shape as input.
        alpha: Regularization factor ([0,1]) between the hinge and the KR loss.
        min_margin: Minimal margin for the hinge loss.
        multi_gpu (bool): set to True when running on multi-GPU/TPU

    Returns:
        The regularized Wasserstein-1 loss.

    See Also:
        :py:func:`hinge_margin_loss`
        :py:func:`kr_loss`
    """
    kr_loss_fct = kr_loss_multi_gpu if multi_gpu else kr_loss
    assert alpha <= 1.0
    if alpha == 1.0:  # alpha for  hinge only
        return hinge_margin_loss(input, target, min_margin)
    if alpha == 0:
        return -kr_loss_fct(input, target)
    # true value: positive value should be the first to be coherent with the
    # hinge loss (positive y_pred)
    return alpha * hinge_margin_loss(input, target, min_margin) - (
        1 - alpha
    ) * kr_loss_fct(input, target)


def hinge_multiclass_loss(
    input: torch.Tensor,
    target: torch.Tensor,
    min_margin: float = 1,
) -> torch.Tensor:
    """
    Loss to estimate the Hinge loss in a multiclass setup. It compute the
    elementwise hinge term. Note that this formulation differs from the
    one commonly found in tensorflow/pytorch (with maximise the difference
    between the two largest logits). This formulation is consistent with the
    binary classification loss used in a multiclass fashion.

    Args:
        input: Tensor of arbitrary shape.
        target: Tensor of the same shape as input containing
            one hot encoding target labels (0 and +1).
        min_margin: The minimal margin to enforce.
    Note:
        target should be one hot encoded. labels in (1,0)

    Returns:
        The hinge margin multiclass loss.
    """
    sign_target = torch.where(target > 0, 1.0, -1.0).to(input.dtype)
    hinge = F.relu(min_margin / 2.0 - sign_target * input)
    # reweight positive elements
    factor = target.shape[-1] - 1.0
    hinge = torch.where(target > 0, hinge * factor, hinge)
    return torch.mean(hinge, dim=-1)


def hkr_multiclass_loss(
    input: torch.Tensor,
    target: torch.Tensor,
    alpha: float = 0.0,
    min_margin: float = 1.0,
    multi_gpu=False,
) -> torch.Tensor:
    """
    Loss to estimate the wasserstein-1 distance with a hinge regularization using
    Kantorovich-Rubinstein duality.

    Args:
        input: Tensor of arbitrary shape.
        target: Tensor of the same shape as input.
        alpha: Regularization factor ([0,1]) between the hinge and the KR loss.
        min_margin: Minimal margin for the hinge loss.
        multi_gpu (bool): set to True when running on multi-GPU/TPU

    Returns:
        The regularized Wasserstein-1 loss.

    See Also:
        :py:func:`hinge_margin_loss`
        :py:func:`kr_loss`
    """

    assert alpha <= 1.0
    kr_loss_fct = kr_loss_multi_gpu if multi_gpu else kr_loss
    if alpha == 1.0:  # alpha  hinge only
        return hinge_multiclass_loss(input, target, min_margin)
    elif alpha == 0.0:  # alpha = 0 => KR only
        return -kr_loss_fct(input, target)
    else:
        return alpha * hinge_multiclass_loss(input, target, min_margin) - (
            1 - alpha
        ) * kr_loss_fct(input, target)


def process_labels_for_multi_gpu(labels: torch.Tensor) -> torch.Tensor:
    """Process labels to be fed to any loss based on KR estimation with a multi-GPU/TPU
    strategy.

    When using a multi-GPU/TPU strategy, the flag `multi_gpu` in KR-based losses must be
    set to True and the labels have to be pre-processed with this function.

    For binary classification, the labels should be of shape [batch_size, 1].
    For multiclass problems, the labels must be one-hot encoded (1 or 0) with shape
    [batch_size, number of classes].

    Args:
        labels (torch.Tensor): tensor containing the labels

    Returns:
        torch.Tensor: labels processed for KR-based losses with multi-GPU/TPU strategy.
    """
    pos_labels = torch.where(labels > 0, 1.0, 0.0).to(labels.dtype)
    mean_pos = torch.mean(pos_labels, dim=0)
    # pos factor = batch_size/number of positive samples
    pos_factor = torch.nan_to_num(1.0 / mean_pos)
    # neg factor = batch_size/number of negative samples
    neg_factor = -torch.nan_to_num(1.0 / (1.0 - mean_pos))

    # Since element-wise KR terms are averaged by loss reduction later on, it is needed
    # to multiply by batch_size here.
    return torch.where(labels > 0, pos_factor, neg_factor)


class SymmetricPad(torch.nn.Module):
    """
    Pads a 2D tensor symmetrically.

    Args:
        pad (tuple): A tuple (pad_left, pad_right, pad_top, pad_bottom) specifying
                 the number of pixels to pad on each side. (or single int if
                 common padding).

        onedim: False for conv2d, True for conv1d.

    """

    def __init__(self, pad, onedim=False):
        super().__init__()
        self.onedim = onedim
        num_dim = 2 if onedim else 4
        if isinstance(pad, int):
            self.pad = (pad,) * num_dim
        else:
            self.pad = torch.nn.modules.utils._reverse_repeat_tuple(pad, 2)
        assert len(self.pad) == num_dim, f"Pad must be a tuple of {num_dim} integers"

    def forward(self, x):
        # Horizontal padding
        left = x[:, ..., : self.pad[0]].flip(dims=[-1])
        right = x[:, ..., -self.pad[1] :].flip(dims=[-1])
        x = torch.cat([left, x, right], dim=-1)
        if self.onedim:
            return x
        # Vertical padding
        top = x[:, :, : self.pad[2], :].flip(dims=[-2])
        bottom = x[:, :, -self.pad[3] :, :].flip(dims=[-2])
        x = torch.cat([top, x, bottom], dim=-2)

        return x



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
import pytest
from functools import partial

import numpy as np

from . import utils_framework as uft

from .utils_framework import (
    tLinear,
    tConv2d,
    bjorck_normalization,
    reshaped_kernel_orthogonalization,
    spectral_normalization,
    spectral_normalization_conv,
    DEFAULT_EPS_SPECTRAL,
)

from .utils_framework import (
    bjorck_norm,
    remove_bjorck_norm,
    frobenius_norm,
    remove_frobenius_norm,
    compute_lconv_coef,
    lconv_norm,
    remove_lconv_norm,
)
from .utils_framework import _padding_circular

rng = np.random.default_rng(42)


@pytest.mark.parametrize(
    "kernel_shape",
    [
        (15, 32),
        (32, 15),
    ],
)
def test_kernel_svd(kernel_shape):
    """Compare max singular value using power iteration and np.linalg.svd"""
    kernel = rng.normal(size=kernel_shape).astype("float32")
    sigmas_svd = np.linalg.svd(
        np.reshape(kernel, (np.prod(kernel.shape[:-1]), kernel.shape[-1])),
        full_matrices=False,
        compute_uv=False,
    )
    SVmax = np.max(sigmas_svd)

    u = rng.normal(size=(1, kernel.shape[-1])).astype("float32")

    kernel = uft.to_tensor(kernel)
    u = uft.to_tensor(u)
    W_bar, _u, sigma = uft.get_instance_framework(
        spectral_normalization,
        {"kernel": kernel, "u": u, "eps": DEFAULT_EPS_SPECTRAL},
    )
    # Test sigma is close to the one computed with svd first run @ 1e-1
    np.testing.assert_approx_equal(
        sigma, SVmax, 1, "test failed with kernel_shape " + str(kernel.shape)
    )

    W_bar, _u, sigma = uft.get_instance_framework(
        spectral_normalization,
        {"kernel": kernel, "u": _u, "eps": DEFAULT_EPS_SPECTRAL},
    )
    # spectral_normalization(kernel, u=_u, eps=1e-6)
    W_bar, _u, sigma = uft.get_instance_framework(
        spectral_normalization,
        {"kernel": kernel, "u": _u, "eps": DEFAULT_EPS_SPECTRAL},
    )
    # spectral_normalization(kernel, u=_u, eps=1e-6)
    # Test W_bar is reshaped correctly
    np.testing.assert_equal(W_bar.shape, (np.prod(kernel.shape[:-1]), kernel.shape[-1]))
    # Test sigma is close to the one computed with svd second run @ 1e-5
    np.testing.assert_approx_equal(
        sigma, SVmax, 3, "test failed with kernel_shape " + str(kernel.shape)
    )
    # Test if kernel is normalized by sigma
    np.testing.assert_allclose(
        np.reshape(W_bar, kernel.shape), kernel / sigmas_svd[0], atol=1e-2
    )


def set_spectral_input_shape(kernel, strides):
    """Set spectral input shape and RO_case, depending on kernel shape and
    strides."""
    kh, kw, c_in, c_out = kernel.shape
    cPad = [kh // 2, kw // 2]
    stride = strides[0]

    # Compute minimal N
    r = kh // 2
    if r < 1:
        N = 5
    else:
        N = 4 * r + 1
        if stride > 1:
            N = int(0.5 + N / stride)

    if c_in * stride**2 > c_out:
        spectral_input_shape = (N, N, c_out)
        RO_case = True
    else:
        spectral_input_shape = (stride * N, stride * N, c_in)
        RO_case = False
    return spectral_input_shape, RO_case, cPad


@pytest.mark.parametrize(
    "kernel_shape, strides",
    [
        ((5, 5, 32, 64), (1, 1)),
        ((3, 3, 12, 8), (1, 1)),
        ((3, 3, 24, 24), (1, 1)),
    ],
)
def test_kernel_conv_svd(kernel_shape, strides):
    """Compare power iteration conv against SVD."""

    if hasattr(spectral_normalization_conv, "unavailable_class"):
        pytest.skip("spectral_normalization_conv not implemented")
    if hasattr(_padding_circular, "unavailable_class"):
        pytest.skip("_padding_circular not implemented")

    np.random.seed(42)

    kernel = np.random.normal(size=kernel_shape).astype("float32")
    spectral_input_shape, RO_case, cPad = set_spectral_input_shape(kernel, strides)

    # Compute max singular value using FFT2 and SVD
    kernel_n = kernel.astype(dtype="float32")
    transforms = np.fft.fft2(
        kernel_n,
        (spectral_input_shape[0], spectral_input_shape[1]),
        axes=[0, 1],
    )
    svd = np.linalg.svd(transforms, compute_uv=False)
    SVmax = np.max(svd)

    # Compute max singular value using power iteration conv
    _u = np.random.normal(size=(1,) + spectral_input_shape).astype("float32")
    fPad = partial(_padding_circular, circular_paddings=cPad)

    kernel = uft.to_tensor(kernel)
    _u = uft.to_tensor(_u)
    res = uft.get_instance_framework(
        spectral_normalization_conv,
        {
            "kernel": kernel,
            "u": _u,
            "stride": strides[0],
            "conv_first": not RO_case,
            "pad_func": fPad,
            "eps": 1e-6,
            "maxiter": 30,
        },
    )
    if res is None:
        pytest.skip("spectral_normalization_conv with params not available")
    W_bar, _u, sigma = res

    # Test if sigma is close to the one computed with svd first run @ 1e-1
    np.testing.assert_approx_equal(
        sigma, SVmax, 1, "test failed with kernel_shape " + str(kernel.shape)
    )

    # Run a second time power iteration conv with last _u from first run
    W_bar, _u, sigma = uft.get_instance_framework(
        spectral_normalization_conv,
        {
            "kernel": kernel,
            "u": _u,
            "stride": strides[0],
            "conv_first": not RO_case,
            "pad_func": fPad,
            "eps": 1e-6,
            "maxiter": 30,
        },
    )
    # Test if W_bar is reshaped correctly
    np.testing.assert_equal(W_bar.shape, kernel.shape)
    # Test if sigma is close to the one computed with svd, second run
    np.testing.assert_approx_equal(
        sigma, SVmax, 2, "test failed with kernel_shape " + str(kernel.shape)
    )
    # Test if kernel is normalized by sigma
    np.testing.assert_allclose(
        np.reshape(W_bar, kernel.shape), kernel / SVmax, atol=1e-2
    )


@pytest.mark.parametrize(
    "kernel_shape",
    [
        (15, 32),
        (64, 32),
    ],
)
def test_bjorck_normalization(kernel_shape):
    np.random.seed(42)

    kernel = np.random.normal(size=kernel_shape).astype("float32")
    """Compare max singular value using power iteration and tf.linalg.svd"""
    sigmas_svd = np.linalg.svd(
        np.reshape(kernel, (np.prod(kernel.shape[:-1]), kernel.shape[-1])),
        full_matrices=False,
        compute_uv=False,
    )
    SVmax = np.max(sigmas_svd)

    kernel = uft.to_tensor(kernel)
    wbar = uft.get_instance_framework(
        bjorck_normalization, {"w": kernel / SVmax, "eps": 1e-5}
    )
    # wbar = bjorck_normalization(kernel / SVmax, eps=1e-5)
    sigmas_wbar_svd = np.linalg.svd(
        np.reshape(wbar, (np.prod(wbar.shape[:-1]), wbar.shape[-1])),
        full_matrices=False,
        compute_uv=False,
    )
    # Test sigma is close to the one computed with svd first run @ 1e-1
    np.testing.assert_allclose(
        sigmas_wbar_svd, np.ones(sigmas_wbar_svd.shape), atol=1e-2
    )
    # Test W_bar is reshaped correctly
    np.testing.assert_equal(wbar.shape, (np.prod(kernel.shape[:-1]), kernel.shape[-1]))

    # Test sigma is close to the one computed with svd second run @1e-5
    wbar = uft.get_instance_framework(bjorck_normalization, {"w": wbar, "eps": 1e-5})
    # wbar = bjorck_normalization(wbar, eps=1e-5)
    sigmas_wbar_svd = np.linalg.svd(
        np.reshape(wbar, (np.prod(wbar.shape[:-1]), wbar.shape[-1])),
        full_matrices=False,
        compute_uv=False,
    )
    np.testing.assert_allclose(
        sigmas_wbar_svd, np.ones(sigmas_wbar_svd.shape), atol=1e-4
    )


@pytest.mark.parametrize(
    "kernel_shape",
    [
        (15, 32),
        (5, 5, 64, 32),
    ],
)
def test_reshaped_kernel_orthogonalization(kernel_shape):
    if hasattr(reshaped_kernel_orthogonalization, "unavailable_class"):
        pytest.skip("reshaped_kernel_orthogonalization not implemented")
    np.random.seed(42)

    kernel = np.random.normal(size=kernel_shape).astype("float32")
    """Compare max singular value using power iteration and tf.linalg.svd"""
    sigmas_svd = np.linalg.svd(
        np.reshape(kernel, (np.prod(kernel.shape[:-1]), kernel.shape[-1])),
        full_matrices=False,
        compute_uv=False,
    )
    SVmax = np.max(sigmas_svd)

    kernel = uft.to_tensor(kernel)
    res = uft.get_instance_framework(
        reshaped_kernel_orthogonalization,
        {
            "kernel": kernel,
            "u": None,
            "adjustment_coef": 1.0,
            "eps_spectral": 1e-5,
            "eps_bjorck": 1e-5,
        },
    )
    if res is None:
        pytest.skip("reshaped_kernel_orthogonalization with params not available")
    W_bar, _, sigma = res
    # Test W_bar is reshaped correctly
    np.testing.assert_equal(W_bar.shape, kernel.shape)
    # Test RKO sigma is close to max(svd)
    np.testing.assert_approx_equal(
        sigma, SVmax, 1, "test failed with kernel_shape " + str(kernel.shape)
    )
    sigmas_wbar_svd = np.linalg.svd(
        np.reshape(W_bar, (np.prod(W_bar.shape[:-1]), W_bar.shape[-1])),
        full_matrices=False,
        compute_uv=False,
    )
    # Test if SVs of W_bar are close to one
    np.testing.assert_allclose(
        sigmas_wbar_svd, np.ones(sigmas_wbar_svd.shape), atol=1e-2
    )


@pytest.mark.skipif(
    hasattr(bjorck_norm, "unavailable_class"),
    reason="bjorck_norm not available",
)
def test_bjorck_norm():
    """
    test bjorck_norm parametrization implementation
    """
    np.random.seed(42)

    m = uft.get_instance_framework(
        tLinear, {"in_features": 2, "out_features": 2}
    )  # torch.nn.Linear(2, 2)
    # torch.nn.init.orthogonal_(m.weight)
    w1 = uft.get_instance_framework(
        bjorck_normalization, {"w": uft.get_layer_weights(m), "eps": 1e-5}
    )
    # bjorck_normalization(m.weight)

    # bjorck norm parametrization
    uft.get_instance_framework(bjorck_norm, {"module": m})  # (m)

    # ensure that the original weight is the only torch parameter
    uft.check_parametrization(m, is_parametrized=True)
    # assert isinstance(m.parametrizations.weight.original, torch.nn.Parameter)
    # assert not isinstance(m.weight, torch.nn.Parameter)

    # check the orthogonality of the weight
    x = np.random.rand(2)
    x = uft.to_tensor(x)
    m(x)
    np.testing.assert_allclose(
        uft.to_numpy(w1), uft.to_numpy(uft.get_layer_weights(m)), atol=1e-5
    )

    # remove the parametrization
    uft.get_instance_framework(remove_bjorck_norm, {"module": m})  # (m)
    uft.check_parametrization(m, is_parametrized=False)
    # assert not hasattr(m, "parametrizations")
    # assert isinstance(m.weight, torch.nn.Parameter)
    np.testing.assert_allclose(
        uft.to_numpy(w1), uft.to_numpy(uft.get_layer_weights(m)), atol=1e-5
    )


@pytest.mark.skipif(
    hasattr(frobenius_norm, "unavailable_class"),
    reason="frobenius_norm not available",
)
def test_frobenius_norm():
    """
    test frobenius_norm parametrization implementation
    """
    np.random.seed(42)

    m = uft.get_instance_framework(
        tLinear, {"in_features": 2, "out_features": 2}
    )  # torch.nn.Linear(2, 2)
    # torch.nn.init.uniform_(m.weight)
    w1 = uft.to_numpy(uft.get_layer_weights(m))
    w1 = w1 / np.linalg.norm(w1)  # m.weight / torch.norm(m.weight)

    # frobenius norm parametrization
    uft.get_instance_framework(frobenius_norm, {"module": m, "disjoint_neurons": False})

    # ensure that the original weight is the only torch parameter
    uft.check_parametrization(m, is_parametrized=True)
    # assert isinstance(m.parametrizations.weight.original, torch.nn.Parameter)
    # assert not isinstance(m.weight, torch.nn.Parameter)

    # check the orthogonality of the weight
    x = np.random.rand(2)
    x = uft.to_tensor(x)
    m(x)
    np.testing.assert_allclose(
        uft.to_numpy(w1), uft.to_numpy(uft.get_layer_weights(m)), atol=1e-5
    )

    # remove the parametrization
    uft.get_instance_framework(remove_frobenius_norm, {"module": m})
    uft.check_parametrization(m, is_parametrized=False)
    np.testing.assert_allclose(
        uft.to_numpy(w1), uft.to_numpy(uft.get_layer_weights(m)), atol=1e-5
    )


@pytest.mark.skipif(
    hasattr(frobenius_norm, "unavailable_class"),
    reason="frobenius_norm not available",
)
def test_frobenius_norm_disjoint_neurons():
    """
    Test `disjoint_neurons=True` argument in frobenius_norm parametrization
    """
    np.random.seed(42)

    params = {"in_features": 5, "out_features": 3}
    m = uft.get_instance_framework(tLinear, params)

    # Set parametrization and perform a forward pass to compute new weights
    uft.get_instance_framework(frobenius_norm, {"module": m, "disjoint_neurons": True})

    x = np.random.rand(5)
    x = uft.to_tensor(x)
    m(x)

    # Assert that all rows of matrix weight are independently normalized
    ww = uft.to_numpy(uft.get_layer_weights(m))
    for i in range(params["out_features"]):
        np.testing.assert_allclose(np.linalg.norm(ww[i, :]), 1.0, rtol=2e-7)


@pytest.mark.skipif(
    hasattr(lconv_norm, "unavailable_class"),
    reason="lconv_norm not available",
)
def test_lconv_norm():
    """
    test lconv_norm parametrization implementation
    """
    np.random.seed(42)

    params = {
        "in_channels": 1,
        "out_channels": 2,
        "kernel_size": (3, 3),
        "stride": (1, 1),
    }
    m = uft.get_instance_framework(tConv2d, params)
    # torch.nn.init.orthogonal_(m.weight)
    w1 = uft.get_layer_weights(m) * compute_lconv_coef(
        params["kernel_size"], None, params["stride"]
    )

    # lconv norm parametrization
    uft.get_instance_framework(lconv_norm, {"module": m})

    shape = [1, 1, 5, 5]
    shape = uft.to_framework_channel(shape)
    x = np.random.rand(*shape)
    x = uft.to_tensor(x)
    _ = m(x)

    # ensure that the original weight is the only torch parameter
    uft.check_parametrization(m, is_parametrized=True)
    # assert isinstance(m.parametrizations.weight.original, torch.nn.Parameter)
    # assert not isinstance(m.weight, torch.nn.Parameter)

    # check the normalization of the weight
    np.testing.assert_allclose(
        uft.to_numpy(w1), uft.to_numpy(uft.get_layer_weights(m)), atol=1e-7
    )
    # tt.assert_equal(y, torch.nn.functional.conv2d(x, w1, bias=m.bias, stride=(1, 1)))

    # remove the parametrization
    uft.get_instance_framework(remove_lconv_norm, {"module": m})
    uft.check_parametrization(m, is_parametrized=False)
    # assert isinstance(m.weight, torch.nn.Parameter)
    np.testing.assert_allclose(
        uft.to_numpy(w1), uft.to_numpy(uft.get_layer_weights(m)), atol=1e-7
    )



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
import pytest
import os
import pprint
import tempfile

import numpy as np


from .utils_framework import (
    SpectralLinear,
    SpectralConv2d,
    SpectralConv1d,
    SpectralConvTranspose2d,
    FrobeniusLinear,
    FrobeniusConv2d,
    ScaledAvgPool2d,
    ScaledAdaptiveAvgPool2d,
    ScaledL2NormPool2d,
    InvertibleDownSampling,
    InvertibleUpSampling,
    ScaledAdaptativeL2NormPool2d,
    Flatten,
    Sequential,
)
from . import utils_framework as uft


from .utils_framework import (
    tLinear,
    AutoWeightClipConstraint,
    SpectralConstraint,
    FrobeniusConstraint,
    tInput,
    CondenseCallback,
    MonitorCallback,
)

pp = pprint.PrettyPrinter(indent=4)

"""
About these tests:
==================

What is tested:
---------------
- layer instantiation
- training
- prediction
- storing on disk and reloading
- k lip_constraint is respected ( at +-0.001 )

What is not tested:
-------------------
- layer performance ( time / accuracy )
- layer structure ( don't check that SpectralConv2d is actually a convolution )

However, all run generate log that can be manually checked with tensorboard
"""


def linear_generator(batch_size, input_shape: tuple, kernel):
    """
    Generate data according to a linear kernel
    Args:
        batch_size: size of each batch
        input_shape: shape of the desired input
        kernel: kernel used to generate data, must match the last dimensions of
            `input_shape`

    Returns:
        a generator for the data

    """
    input_shape = tuple(input_shape)
    while True:
        # pick random sample in [0, 1] with the input shape
        batch_x = np.array(
            np.random.uniform(-10, 10, (batch_size,) + input_shape), dtype=np.float16
        )
        # apply the k lip linear transformation
        batch_y = np.tensordot(
            batch_x,
            kernel,
            axes=(
                [i for i in range(1, len(input_shape) + 1)],
                [i for i in range(0, len(input_shape))],
            ),
        )
        yield batch_x, batch_y


def build_kernel(input_shape: tuple, output_shape: tuple, k=1.0):
    """
    build a kernel with defined lipschitz factor

    Args:
        input_shape: input shape of the linear function
        output_shape: output shape of the linear function
        k: lipshitz factor of the function

    Returns:
        the kernel for use in the linear_generator

    """
    input_shape = tuple(input_shape)
    output_shape = tuple(output_shape)
    kernel = np.array(
        np.random.random_sample(input_shape + output_shape), dtype=np.float16
    )
    kernel = (
        kernel * k / np.linalg.norm(kernel)
    )  # assuming lipschitz constraint is independent with respect to the chosen metric

    return kernel


def build_and_train_k_lip_model(
    layer_type: type,
    layer_params: dict,
    batch_size: int,
    epochs: int,
    input_shape: tuple,
    k_lip_model: float,
    k_lip_data: float,
    **kwargs,
):
    # clear session to avoid side effects from previous train
    uft.init_session()  # K.clear_session()
    np.random.seed(42)
    input_shape = uft.to_framework_channel(input_shape)
    # create the model, defin opt, and compile it
    model = uft.generate_k_lip_model(layer_type, layer_params, input_shape, k_lip_model)

    optimizer = uft.get_instance_framework(
        uft.Adam, inst_params={"lr": 0.001, "model": model}
    )

    loss_fn, optimizer, metrics = uft.compile_model(
        model,
        optimizer=optimizer,
        loss=uft.MeanSquaredError(),
        metrics=[uft.metric_mse()],
    )
    # model.compile(optimizer=optimizer, loss="mean_squared_error", )
    # create the synthetic data generator
    output_shape = uft.compute_output_shape(input_shape, model)
    kernel = build_kernel(input_shape, output_shape, k_lip_data)
    # define logging features
    logdir = os.path.join("logs", uft.LIP_LAYERS, "%s" % layer_type.__name__)
    os.makedirs(logdir, exist_ok=True)

    callback_list = []
    if kwargs["callbacks"] is not None:
        callback_list = callback_list + kwargs["callbacks"]
    # train model

    traind_ds = linear_generator(batch_size, input_shape, kernel)
    uft.train(
        traind_ds,
        model,
        loss_fn,
        optimizer,
        epochs,
        batch_size,
        steps_per_epoch=10,
    )
    return model, kernel, loss_fn, metrics


def train_k_lip_model(
    layer_type: type,
    layer_params: dict,
    batch_size: int,
    steps_per_epoch: int,
    epochs: int,
    input_shape: tuple,
    k_lip_model: float,
    k_lip_data: float,
    **kwargs,
):
    """
    Create a generator, create a model, train it and return the results.

    Args:
        layer_type:
        layer_params:
        batch_size:
        steps_per_epoch:
        epochs:
        input_shape:
        k_lip_model:
        k_lip_data:
        **kwargs:

    Returns:
        the generator

    """
    # define logging features
    logdir = os.path.join("logs", uft.LIP_LAYERS, "%s" % layer_type.__name__)
    os.makedirs(logdir, exist_ok=True)

    model, kernel, loss_fn, metrics = build_and_train_k_lip_model(
        layer_type=layer_type,
        layer_params=layer_params,
        batch_size=batch_size,
        epochs=epochs,
        input_shape=input_shape,
        k_lip_model=k_lip_model,
        k_lip_data=k_lip_data,
        **kwargs,
    )
    # the seed is set to compare all models with the same data
    test_dl = linear_generator(batch_size, input_shape, kernel)
    np.random.seed(42)
    uft.set_seed(42)
    x, y = test_dl.send(None)

    loss, mse = uft.run_test(model, test_dl, loss_fn, metrics, steps=10)

    x = uft.to_tensor(x)
    empirical_lip_const = uft.evaluate_lip_const(model=model, x=x, seed=42)
    # save the model
    model_checkpoint_path = os.path.join(logdir, uft.MODEL_PATH)
    uft.save_model(model, model_checkpoint_path, overwrite=True)
    # model.save(model_checkpoint_path, overwrite=True)
    del model
    uft.init_session()  # K.clear_session()
    model = uft.load_model(
        model_checkpoint_path,
        layer_type=layer_type,
        layer_params=layer_params,
        input_shape=input_shape,
        k=k_lip_model,
    )
    np.random.seed(42)
    uft.set_seed(42)
    test_dl = linear_generator(batch_size, input_shape, kernel)  # .send(None)
    x, y = test_dl.send(None)
    from_disk_loss, from_disk_mse = uft.run_test(
        model, test_dl, loss_fn, metrics, steps=10
    )
    x = uft.to_tensor(x)
    from_empirical_lip_const = uft.evaluate_lip_const(model=model, x=x, seed=42)

    # log metrics
    return (
        mse,
        uft.to_numpy(empirical_lip_const),
        from_disk_mse,
        uft.to_numpy(from_empirical_lip_const),
    )


def _check_mse_results(mse, from_disk_mse, test_params):
    assert from_disk_mse == pytest.approx(
        mse, 1e-5
    ), "serialization must not change the performance of a layer"


def _check_emp_lip_const(emp_lip_const, from_disk_emp_lip_const, test_params):
    assert from_disk_emp_lip_const == pytest.approx(
        emp_lip_const, 1e-5
    ), "serialization must not change the Lipschitz constant of a layer"
    assert (
        emp_lip_const <= test_params["k_lip_model"] * 1.02
    ), " the lip const of the network must be lower than the specified boundary"


def _apply_tests_bank(test_params):
    pp.pprint(test_params)
    (
        mse,
        emp_lip_const,
        from_disk_mse,
        from_disk_emp_lip_const,
    ) = train_k_lip_model(**test_params)
    print("test mse: %f" % mse)
    print(
        "empirical lip const: %f ( expected min data and model %s )"
        % (
            emp_lip_const,
            min(test_params["k_lip_model"], test_params["k_lip_data"]),
        )
    )
    _check_mse_results(mse, from_disk_mse, test_params)
    _check_emp_lip_const(emp_lip_const, from_disk_emp_lip_const, test_params)


@pytest.mark.skipif(
    hasattr(AutoWeightClipConstraint, "unavailable_class"),
    reason="AutoWeightClipConstraint not available",
)
@pytest.mark.parametrize(
    "test_params",
    [
        dict(
            layer_type=tLinear,
            layer_params={
                "in_features": 4,
                "out_features": 4,
                "kernel_constraint": AutoWeightClipConstraint(1.0),
            },
            batch_size=250,
            steps_per_epoch=125,
            epochs=5,
            input_shape=(4,),
            k_lip_data=1.0,
            k_lip_model=1.0,
            callbacks=[],
        ),
        dict(
            layer_type=tLinear,
            layer_params={
                "in_features": 4,
                "out_features": 4,
                "kernel_constraint": AutoWeightClipConstraint(1.0),
            },
            batch_size=250,
            steps_per_epoch=125,
            epochs=5,
            input_shape=(4,),
            k_lip_data=5.0,
            k_lip_model=1.0,
            callbacks=[],
        ),
        dict(
            layer_type=tLinear,
            layer_params={
                "in_features": 4,
                "out_features": 4,
                "kernel_constraint": AutoWeightClipConstraint(5.0),
            },
            batch_size=250,
            steps_per_epoch=125,
            epochs=5,
            input_shape=(4,),
            k_lip_data=1.0,
            k_lip_model=5.0,
            callbacks=[],
        ),
    ],
)
def test_constraints_clipping(test_params):
    """
    Tests for a standard Linear layer, for result comparison.
    """
    _apply_tests_bank(test_params)


@pytest.mark.skipif(
    hasattr(SpectralConstraint, "unavailable_class"),
    reason="SpectralConstraint not available",
)
@pytest.mark.parametrize(
    "test_params",
    [
        dict(
            layer_type=tLinear,
            layer_params={
                "in_features": 4,
                "out_features": 4,
                "kernel_constraint": SpectralConstraint(1.0),
            },
            batch_size=250,
            steps_per_epoch=125,
            epochs=5,
            input_shape=(4,),
            k_lip_data=1.0,
            k_lip_model=1.0,
            callbacks=[],
        ),
        dict(
            layer_type=tLinear,
            layer_params={
                "in_features": 4,
                "out_features": 4,
                "kernel_constraint": SpectralConstraint(1.0),
            },
            batch_size=250,
            steps_per_epoch=125,
            epochs=5,
            input_shape=(4,),
            k_lip_data=5.0,
            k_lip_model=1.0,
            callbacks=[],
        ),
        dict(
            layer_type=tLinear,
            layer_params={
                "in_features": 4,
                "out_features": 4,
                "kernel_constraint": SpectralConstraint(5.0),
            },
            batch_size=250,
            steps_per_epoch=125,
            epochs=5,
            input_shape=(4,),
            k_lip_data=1.0,
            k_lip_model=5.0,
            callbacks=[],
        ),
    ],
)
def test_constraints_orthogonal(test_params):
    """
    Tests for a standard Linear layer, for result comparison.
    """
    _apply_tests_bank(test_params)


@pytest.mark.skipif(
    hasattr(FrobeniusConstraint, "unavailable_class"),
    reason="FrobeniusConstraint not available",
)
@pytest.mark.parametrize(
    "test_params",
    [
        dict(
            layer_type=tLinear,
            layer_params={
                "in_features": 4,
                "out_features": 4,
                "kernel_constraint": FrobeniusConstraint(),
            },
            batch_size=250,
            steps_per_epoch=125,
            epochs=5,
            input_shape=(4,),
            k_lip_data=1.0,
            k_lip_model=1.0,
            callbacks=[],
        ),
        dict(
            layer_type=tLinear,
            layer_params={
                "in_features": 4,
                "out_features": 4,
                "kernel_constraint": FrobeniusConstraint(),
            },
            batch_size=250,
            steps_per_epoch=125,
            epochs=5,
            input_shape=(4,),
            k_lip_data=5.0,
            k_lip_model=1.0,
            callbacks=[],
        ),
    ],
)
def test_constraints_frobenius(test_params):
    """
    Tests for a standard Linear layer, for result comparison.
    """
    _apply_tests_bank(test_params)


@pytest.mark.parametrize(
    "layer_type",
    [
        SpectralLinear,
    ],
)
@pytest.mark.parametrize(
    "layer_params,k_lip_data,k_lip_model",
    [
        (
            {"bias": False, "in_features": 4, "out_features": 3},
            1.0,
            1.0,
        ),
        (
            {"in_features": 4, "out_features": 4},
            5.0,
            1.0,
        ),
        (
            {"in_features": 4, "out_features": 4},
            1.0,
            5.0,
        ),
    ],
)
def test_spectral_linear(layer_type, layer_params, k_lip_data, k_lip_model):
    test_params = dict(
        layer_type=layer_type,
        layer_params=layer_params,
        batch_size=250,
        steps_per_epoch=125,
        epochs=5,
        input_shape=(4,),
        k_lip_data=k_lip_data,
        k_lip_model=k_lip_model,
        callbacks=[],
    )
    _apply_tests_bank(test_params)


@pytest.mark.parametrize(
    "layer_type",
    [
        FrobeniusLinear,
    ],
)
@pytest.mark.parametrize(
    "layer_params,k_lip_data,k_lip_model",
    [
        (
            {"bias": False, "in_features": 4, "out_features": 1},
            1.0,
            1.0,
        ),
        (
            {"in_features": 4, "out_features": 1},
            5.0,
            1.0,
        ),
        (
            {"in_features": 4, "out_features": 1},
            1.0,
            5.0,
        ),
    ],
)
def test_frobenius_linear(layer_type, layer_params, k_lip_data, k_lip_model):
    test_params = dict(
        layer_type=layer_type,
        layer_params=layer_params,
        batch_size=250,
        steps_per_epoch=125,
        epochs=5,
        input_shape=(4,),
        k_lip_data=k_lip_data,
        k_lip_model=k_lip_model,
        callbacks=[],
    )
    _apply_tests_bank(test_params)


@pytest.mark.parametrize(
    "layer_type",
    [SpectralConv2d, FrobeniusConv2d, SpectralConvTranspose2d],
)
@pytest.mark.parametrize(
    "layer_params,k_lip_data,k_lip_model",
    [
        (
            {
                "in_channels": 1,
                "out_channels": 2,
                "kernel_size": (3, 3),
                "bias": False,
            },
            1.0,
            1.0,
        ),
        (
            {"in_channels": 1, "out_channels": 2, "kernel_size": (3, 3)},
            5.0,
            1.0,
        ),
        (
            {"in_channels": 1, "out_channels": 2, "kernel_size": (3, 3)},
            1.0,
            5.0,
        ),
    ],
)
def test_conv2d(layer_type, layer_params, k_lip_data, k_lip_model):
    if hasattr(layer_type, "unavailable_class"):
        pytest.skip("layer not available")
    test_params = dict(
        layer_type=layer_type,
        layer_params=layer_params,
        batch_size=250,
        steps_per_epoch=125,
        epochs=5,
        input_shape=(1, 5, 5),
        k_lip_data=k_lip_data,
        k_lip_model=k_lip_model,
        callbacks=[],
    )
    _apply_tests_bank(test_params)


@pytest.mark.parametrize(
    "pad_mode",
    [
        "zeros",
        "reflect",
        "circular",
        "symmetric",
    ],
)
@pytest.mark.parametrize(
    "pad, kernel_size",
    [
        (1, (3, 3)),
        ((1, 1), (3, 3)),
        (2, (5, 5)),
        ((2, 2), (5, 5)),
    ],
)
@pytest.mark.parametrize(
    "layer_params,k_lip_data,k_lip_model",
    [
        (
            {
                "in_channels": 1,
                "out_channels": 2,
                "bias": False,
            },
            1.0,
            1.0,
        ),
        (
            {"in_channels": 1, "out_channels": 2},
            1.0,
            1.0,
        ),
        (
            {"in_channels": 1, "out_channels": 2},
            1.0,
            5.0,
        ),
    ],
)
def test_spectralconv2d_pad(
    pad, pad_mode, kernel_size, layer_params, k_lip_data, k_lip_model
):
    layer_params["padding"] = pad
    layer_params["padding_mode"] = pad_mode
    layer_params["kernel_size"] = kernel_size
    if not uft.is_supported_padding(pad_mode, SpectralConv2d):
        pytest.skip(f"SpectralConv2d: Padding {pad_mode} not supported")
    test_params = dict(
        layer_type=SpectralConv2d,
        layer_params=layer_params,
        batch_size=250,
        steps_per_epoch=125,
        epochs=5,
        input_shape=(1, 5, 5),
        k_lip_data=k_lip_data,
        k_lip_model=k_lip_model,
        callbacks=[],
    )
    _apply_tests_bank(test_params)


@pytest.mark.skipif(
    hasattr(SpectralConv1d, "unavailable_class"),
    reason="SpectralConv1d not available",
)
@pytest.mark.parametrize(
    "test_params",
    [
        dict(
            layer_type=SpectralConv1d,
            layer_params={
                "in_channels": 1,
                "out_channels": 2,
                "kernel_size": 3,
                "bias": False,
            },
            batch_size=250,
            steps_per_epoch=125,
            epochs=5,
            input_shape=(1, 5),
            k_lip_data=1.0,
            k_lip_model=1.0,
            callbacks=[],
        ),
        dict(
            layer_type=SpectralConv1d,
            layer_params={"in_channels": 1, "out_channels": 2, "kernel_size": 3},
            batch_size=250,
            steps_per_epoch=125,
            epochs=5,
            input_shape=(1, 5),
            k_lip_data=5.0,
            k_lip_model=1.0,
            callbacks=[],
        ),
        dict(
            layer_type=SpectralConv1d,
            layer_params={"in_channels": 1, "out_channels": 2, "kernel_size": 3},
            batch_size=250,
            steps_per_epoch=125,
            epochs=5,
            input_shape=(1, 5),
            k_lip_data=1.0,
            k_lip_model=5.0,
            callbacks=[],
        ),
    ],
)
def test_spectralconv1d(test_params):
    _apply_tests_bank(test_params)


@pytest.mark.parametrize(
    "test_params",
    [
        dict(
            layer_type=ScaledAvgPool2d,
            layer_params={"kernel_size": (3, 3)},
            batch_size=250,
            steps_per_epoch=1,
            epochs=1,
            input_shape=(1, 6, 6),
            k_lip_data=1.0,
            k_lip_model=1.0,
            callbacks=[],
        ),
        dict(
            layer_type=ScaledAvgPool2d,
            layer_params={"kernel_size": (3, 3)},
            batch_size=250,
            steps_per_epoch=1,
            epochs=1,
            input_shape=(1, 6, 6),
            k_lip_data=1.0,
            k_lip_model=5.0,
            callbacks=[],
        ),
        dict(
            layer_type=ScaledAvgPool2d,
            layer_params={"kernel_size": (3, 3)},
            batch_size=250,
            steps_per_epoch=1,
            epochs=1,
            input_shape=(1, 6, 6),
            k_lip_data=5.0,
            k_lip_model=1.0,
            callbacks=[],
        ),
    ],
)
def test_ScaledAvgPool2d(test_params):
    # tests only checks that lip cons is enforced
    _apply_tests_bank(test_params)


@pytest.mark.parametrize(
    "test_params",
    [
        # tests only checks that lip cons is enforced
        dict(
            layer_type=Sequential,
            layer_params={
                "layers": [
                    tInput(uft.to_framework_channel((1, 5, 5))),
                    uft.get_instance_framework(
                        ScaledAdaptiveAvgPool2d,
                        {"data_format": "channels_last", "output_size": (1, 1)},
                    ),
                ]
            },
            batch_size=250,
            steps_per_epoch=1,
            epochs=1,
            input_shape=(1, 5, 5),
            k_lip_data=1.0,
            k_lip_model=1.0,
            callbacks=[],
        ),
        dict(
            layer_type=Sequential,
            layer_params={
                "layers": [
                    tInput(uft.to_framework_channel((1, 5, 5))),
                    uft.get_instance_framework(
                        ScaledAdaptiveAvgPool2d,
                        {"data_format": "channels_last", "output_size": (1, 1)},
                    ),
                ]
            },
            batch_size=250,
            steps_per_epoch=1,
            epochs=1,
            input_shape=(1, 5, 5),
            k_lip_data=5.0,
            k_lip_model=1.0,
            callbacks=[],
        ),
        dict(
            layer_type=Sequential,
            layer_params={
                "layers": [
                    tInput(uft.to_framework_channel((1, 5, 5))),
                    uft.get_instance_framework(
                        ScaledAdaptiveAvgPool2d,
                        {"data_format": "channels_last", "output_size": (1, 1)},
                    ),
                ]
            },
            batch_size=250,
            steps_per_epoch=1,
            epochs=1,
            input_shape=(1, 5, 5),
            k_lip_data=1.0,
            k_lip_model=5.0,
            callbacks=[],
        ),
    ],
)
def test_ScaledAdaptiveAvgPool2d(test_params):
    # tests only checks that lip cons is enforced
    _apply_tests_bank(test_params)


@pytest.mark.parametrize(
    "test_params",
    [
        # tests only checks that lip cons is enforced
        dict(
            layer_type=Sequential,
            layer_params={
                "layers": [
                    tInput(uft.to_framework_channel((1, 5, 5))),
                    uft.get_instance_framework(
                        ScaledL2NormPool2d,
                        {"kernel_size": (2, 3), "data_format": "channels_last"},
                    ),
                ]
            },
            batch_size=250,
            steps_per_epoch=1,
            epochs=1,
            input_shape=(1, 5, 5),
            k_lip_data=1.0,
            k_lip_model=1.0,
            callbacks=[],
        ),
        dict(
            layer_type=Sequential,
            layer_params={
                "layers": [
                    tInput(uft.to_framework_channel((1, 5, 5))),
                    uft.get_instance_framework(
                        ScaledL2NormPool2d,
                        {"kernel_size": (2, 3), "data_format": "channels_last"},
                    ),
                ]
            },
            batch_size=250,
            steps_per_epoch=1,
            epochs=1,
            input_shape=(1, 5, 5),
            k_lip_data=5.0,
            k_lip_model=1.0,
            callbacks=[],
        ),
        dict(
            layer_type=Sequential,
            layer_params={
                "layers": [
                    tInput(uft.to_framework_channel((1, 5, 5))),
                    uft.get_instance_framework(
                        ScaledL2NormPool2d,
                        {"kernel_size": (2, 3), "data_format": "channels_last"},
                    ),
                ]
            },
            batch_size=250,
            steps_per_epoch=1,
            epochs=1,
            input_shape=(1, 5, 5),
            k_lip_data=1.0,
            k_lip_model=5.0,
            callbacks=[],
        ),
    ],
)
def test_scaledl2normPool2d(test_params):
    # tests only checks that lip cons is enforced
    _apply_tests_bank(test_params)


@pytest.mark.skipif(
    hasattr(ScaledAdaptativeL2NormPool2d, "unavailable_class"),
    reason="compute_layer_sv not available",
)
@pytest.mark.parametrize(
    "test_params",
    [
        # tests only checks that lip cons is enforced
        dict(
            layer_type=Sequential,
            layer_params={
                "layers": [
                    tInput(uft.to_framework_channel((1, 5, 5))),
                    uft.get_instance_framework(
                        ScaledAdaptativeL2NormPool2d, {"data_format": "channels_last"}
                    ),
                ]
            },
            batch_size=250,
            steps_per_epoch=1,
            epochs=1,
            input_shape=(1, 5, 5),
            k_lip_data=1.0,
            k_lip_model=1.0,
            callbacks=[],
        ),
        dict(
            layer_type=Sequential,
            layer_params={
                "layers": [
                    tInput(uft.to_framework_channel((1, 5, 5))),
                    uft.get_instance_framework(
                        ScaledAdaptativeL2NormPool2d, {"data_format": "channels_last"}
                    ),
                ]
            },
            batch_size=250,
            steps_per_epoch=1,
            epochs=1,
            input_shape=(1, 5, 5),
            k_lip_data=5.0,
            k_lip_model=1.0,
            callbacks=[],
        ),
        dict(
            layer_type=Sequential,
            layer_params={
                "layers": [
                    tInput(uft.to_framework_channel((1, 5, 5))),
                    uft.get_instance_framework(
                        ScaledAdaptativeL2NormPool2d, {"data_format": "channels_last"}
                    ),
                ]
            },
            batch_size=250,
            steps_per_epoch=1,
            epochs=1,
            input_shape=(1, 5, 5),
            k_lip_data=1.0,
            k_lip_model=5.0,
            callbacks=[],
        ),
    ],
)
def test_scaledgloball2normPool2d(test_params):
    # tests only checks that lip cons is enforced
    _apply_tests_bank(test_params)


@pytest.mark.parametrize(
    "test_params",
    [
        dict(
            layer_type=Sequential,
            layer_params={
                "layers": [
                    tInput(uft.to_framework_channel((1, 5, 5))),
                    uft.get_instance_framework(
                        SpectralConv2d,
                        {
                            "in_channels": 1,
                            "out_channels": 2,
                            "kernel_size": (3, 3),
                            "padding": 1,
                            "bias": False,
                        },
                    ),
                    uft.get_instance_framework(Flatten, {}),
                    uft.get_instance_framework(
                        SpectralLinear, {"in_features": 50, "out_features": 4}
                    ),
                ]
            },
            batch_size=250,
            steps_per_epoch=125,
            epochs=5,
            input_shape=(1, 5, 5),
            k_lip_data=1.0,
            k_lip_model=1.0,
            callbacks=[],
        ),
        dict(
            layer_type=Sequential,
            layer_params={
                "layers": [
                    tInput(uft.to_framework_channel((1, 5, 5))),
                    uft.get_instance_framework(
                        SpectralConv2d,
                        {
                            "in_channels": 1,
                            "out_channels": 2,
                            "padding": 1,
                            "kernel_size": (3, 3),
                        },
                    ),
                    uft.get_instance_framework(Flatten, {}),
                    uft.get_instance_framework(
                        SpectralLinear, {"in_features": 50, "out_features": 4}
                    ),
                ]
            },
            batch_size=250,
            steps_per_epoch=125,
            epochs=5,
            input_shape=(1, 5, 5),
            k_lip_data=5.0,
            k_lip_model=1.0,
            callbacks=[],
        ),
        dict(
            layer_type=Sequential,
            layer_params={
                "layers": [
                    tInput(uft.to_framework_channel((1, 5, 5))),
                    uft.get_instance_framework(
                        SpectralConv2d,
                        {
                            "in_channels": 1,
                            "out_channels": 2,
                            "padding": 1,
                            "kernel_size": (3, 3),
                        },
                    ),
                    uft.get_instance_framework(Flatten, {}),
                    uft.get_instance_framework(
                        SpectralLinear, {"in_features": 50, "out_features": 4}
                    ),
                ]
            },
            batch_size=250,
            steps_per_epoch=125,
            epochs=5,
            input_shape=(1, 5, 5),
            k_lip_data=1.0,
            k_lip_model=5.0,
            callbacks=[],
        ),
    ],
)
def test_multilayer(test_params):
    _apply_tests_bank(test_params)


def build_test_callbacks():
    list_tests = []
    if not hasattr(CondenseCallback, "unavailable_class"):
        list_tests.append(
            dict(
                layer_type=Sequential,
                layer_params={
                    "layers": [
                        tInput(uft.to_framework_channel((1, 5, 5))),
                        uft.get_instance_framework(
                            SpectralConv2d,
                            {
                                "in_channels": 1,
                                "out_channels": 2,
                                "padding": 1,
                                "kernel_size": (3, 3),
                            },
                        ),
                        uft.get_instance_framework(
                            ScaledAvgPool2d, {"kernel_size": (2, 2)}
                        ),
                        uft.get_instance_framework(Flatten, {}),
                        uft.get_instance_framework(
                            SpectralLinear, {"in_features": 8, "out_features": 4}
                        ),
                    ]
                },
                batch_size=250,
                steps_per_epoch=125,
                epochs=5,
                input_shape=(1, 5, 5),
                k_lip_data=1.0,
                k_lip_model=1.0,
                callbacks=[CondenseCallback(on_batch=True, on_epoch=False)],
            )
        )

    if not hasattr(MonitorCallback, "unavailable_class"):
        list_tests.append(
            dict(
                layer_type=Sequential,
                layer_params={
                    "layers": [
                        tInput(uft.to_framework_channel((1, 5, 5))),
                        uft.get_instance_framework(
                            SpectralConv2d,
                            {
                                "in_channels": 1,
                                "out_channels": 2,
                                "kernel_size": (3, 3),
                                "padding": 1,
                                "name": "conv1",
                            },
                        ),
                        uft.get_instance_framework(
                            ScaledAvgPool2d, {"kernel_size": (2, 2)}
                        ),
                        uft.get_instance_framework(Flatten, {}),
                        uft.get_instance_framework(
                            SpectralLinear,
                            {"name": "dense1", "in_features": 8, "out_features": 4},
                        ),
                    ]
                },
                batch_size=250,
                steps_per_epoch=125,
                epochs=5,
                input_shape=(1, 5, 5),
                k_lip_data=1.0,
                k_lip_model=1.0,
                callbacks=[
                    # CondenseCallback(on_batch=False, on_epoch=True),
                    MonitorCallback(
                        monitored_layers=["conv1", "dense1"],
                        logdir=os.path.join("logs", uft.LIP_LAYERS, "Sequential"),
                        target="kernel",
                        what="all",
                        on_epoch=False,
                        on_batch=True,
                    ),
                    MonitorCallback(
                        monitored_layers=["conv1", "dense1"],
                        logdir=os.path.join("logs", uft.LIP_LAYERS, "Sequential"),
                        target="wbar",
                        what="all",
                        on_epoch=False,
                        on_batch=True,
                    ),
                ],
            )
        )
    return list_tests


@pytest.mark.parametrize("test_params", build_test_callbacks())
def test_callbacks(test_params):
    _apply_tests_bank(test_params)


@pytest.mark.parametrize(
    "test_params",
    [
        dict(
            layer_type=InvertibleDownSampling,
            layer_params={"kernel_size": 3},
            batch_size=250,
            steps_per_epoch=1,
            epochs=5,
            input_shape=(3, 6, 6),
            k_lip_data=1.0,
            k_lip_model=1.0,
            callbacks=[],
        ),
    ],
)
def test_invertibledownsampling(test_params):
    # tests only checks that lip cons is enforced
    _apply_tests_bank(test_params)


@pytest.mark.parametrize(
    "test_params",
    [
        dict(
            layer_type=InvertibleUpSampling,
            layer_params={"kernel_size": 3},
            batch_size=250,
            steps_per_epoch=1,
            epochs=5,
            input_shape=(18, 6, 6),
            k_lip_data=1.0,
            k_lip_model=1.0,
            callbacks=[],
        ),
    ],
)
def test_invertibleupsampling(test_params):
    # tests only checks that lip cons is enforced
    _apply_tests_bank(test_params)


@pytest.mark.skipif(
    hasattr(SpectralConvTranspose2d, "unavailable_class"),
    reason="SpectralConvTranspose2d not available",
)
@pytest.mark.parametrize(
    "test_params,msg",
    [
        (dict(in_channels=1, out_channels=5, kernel_size=3), ""),
        (
            dict(in_channels=1, out_channels=12, kernel_size=5, stride=2, bias=False),
            "",
        ),
        (
            dict(
                in_channels=1,
                out_channels=3,
                kernel_size=3,
                padding="same",
                dilation=1,
            ),
            "",
        ),
        (
            dict(
                in_channels=1,
                out_channels=4,
                kernel_size=1,
                output_padding=None,
                activation="relu",
            ),
            "",
        ),
        (
            dict(
                in_channels=1,
                out_channels=16,
                kernel_size=3,
                data_format="channels_first",
            ),
            "",
        ),
        (
            dict(
                in_channels=1,
                out_channels=10,
                kernel_size=3,
                padding=0,
                padding_mode="valid",
            ),
            "Wrong padding",
        ),
        (
            dict(in_channels=1, out_channels=10, kernel_size=3, dilation=2),
            "Wrong dilation rate",
        ),
        (
            dict(in_channels=1, out_channels=10, kernel_size=3, output_padding=5),
            "Wrong data format",
        ),
    ],
)
def test_SpectralConvTranspose2d_instantiation(test_params, msg):
    if msg == "":
        uft.get_instance_framework(SpectralConvTranspose2d, test_params)
    else:
        with pytest.raises(ValueError):
            uft.get_instance_framework(SpectralConvTranspose2d, test_params)


@pytest.mark.skipif(
    hasattr(SpectralConv1d, "unavailable_class"),
    reason="SpectralConv1d not available",
)
@pytest.mark.parametrize(
    "pad_mode",
    [
        "zeros",
        "reflect",
        "circular",
        "symmetric",
    ],
)
@pytest.mark.parametrize(
    "pad, kernel_size",
    [
        (1, (3,)),
        (2, (5,)),
    ],
)
@pytest.mark.parametrize(
    "layer_type",
    [
        SpectralConv1d,
    ],
)
@pytest.mark.parametrize(
    "layer_params",
    [
        {
            "in_channels": 1,
            "out_channels": 2,
            "bias": False,
        },
        {"in_channels": 1, "out_channels": 2},
    ],
)
def test_SpectralConv1d_vanilla_export(
    pad, pad_mode, kernel_size, layer_params, layer_type
):
    layer_params["padding"] = pad
    layer_params["padding_mode"] = pad_mode
    layer_params["kernel_size"] = kernel_size
    layer_type = layer_type
    input_shape = (1, 5)

    model = uft.generate_k_lip_model(layer_type, layer_params, input_shape, 1.0)

    # lay = SpectralConvTranspose2d(**kwargs)
    # model = Sequential([lay])
    x = np.random.normal(size=(5,) + input_shape)

    x = uft.to_tensor(x)
    y1 = model(x)

    # Test vanilla export inference comparison
    if uft.vanilla_require_a_copy():
        model2 = uft.generate_k_lip_model(layer_type, layer_params, input_shape, 1.0)
        uft.copy_model_parameters(model, model2)
        vanilla_model = uft.vanillaModel(model2)
    else:
        vanilla_model = uft.vanillaModel(model)  # .vanilla_export()
    y2 = vanilla_model(x)
    np.testing.assert_allclose(uft.to_numpy(y1), uft.to_numpy(y2), atol=1e-6)

    # Test saving/loading model
    with tempfile.TemporaryDirectory() as tmpdir:
        model_path = os.path.join(tmpdir, uft.MODEL_PATH)
        uft.save_model(model, model_path, overwrite=True)
        uft.load_model(
            model_path,
            layer_type=layer_type,
            layer_params=layer_params,
            input_shape=input_shape,
            k=1.0,
        )


@pytest.mark.skipif(
    hasattr(SpectralConv2d, "unavailable_class"),
    reason="SpectralConv2d not available",
)
@pytest.mark.parametrize(
    "pad_mode",
    [
        "zeros",
        "reflect",
        "circular",
        "symmetric",
    ],
)
@pytest.mark.parametrize(
    "pad, kernel_size",
    [
        (1, (3, 3)),
        ((1, 1), (3, 3)),
        (2, (5, 5)),
        ((2, 2), (5, 5)),
    ],
)
@pytest.mark.parametrize(
    "layer_type",
    [
        SpectralConv2d,
        FrobeniusConv2d,
    ],
)
@pytest.mark.parametrize(
    "layer_params",
    [
        {
            "in_channels": 1,
            "out_channels": 2,
            "bias": False,
        },
        {"in_channels": 1, "out_channels": 2},
    ],
)
def test_Conv2d_vanilla_export(pad, pad_mode, kernel_size, layer_params, layer_type):
    layer_params["padding"] = pad
    layer_params["padding_mode"] = pad_mode
    layer_params["kernel_size"] = kernel_size
    layer_type = layer_type
    input_shape = (1, 5, 5)

    if not uft.is_supported_padding(pad_mode, layer_type):
        pytest.skip(f"{layer_type}: Padding {pad_mode} not supported")
    model = uft.generate_k_lip_model(layer_type, layer_params, input_shape, 1.0)

    # lay = SpectralConvTranspose2d(**kwargs)
    # model = Sequential([lay])
    x = np.random.normal(size=(5,) + input_shape)

    x = uft.to_tensor(x)
    y1 = model(x)

    # Test vanilla export inference comparison
    if uft.vanilla_require_a_copy():
        model2 = uft.generate_k_lip_model(layer_type, layer_params, input_shape, 1.0)
        uft.copy_model_parameters(model, model2)
        vanilla_model = uft.vanillaModel(model2)
    else:
        vanilla_model = uft.vanillaModel(model)  # .vanilla_export()
    y2 = vanilla_model(x)
    np.testing.assert_allclose(uft.to_numpy(y1), uft.to_numpy(y2), atol=1e-6)

    # Test saving/loading model
    with tempfile.TemporaryDirectory() as tmpdir:
        model_path = os.path.join(tmpdir, uft.MODEL_PATH)
        uft.save_model(model, model_path, overwrite=True)
        uft.load_model(
            model_path,
            layer_type=layer_type,
            layer_params=layer_params,
            input_shape=input_shape,
            k=1.0,
        )


@pytest.mark.skipif(
    hasattr(SpectralConvTranspose2d, "unavailable_class"),
    reason="SpectralConvTranspose2d not available",
)
def test_SpectralConvTranspose2d_vanilla_export():
    kwargs = dict(
        in_channels=3,
        out_channels=16,
        kernel_size=5,
        stride=2,
        activation="relu",
        data_format="channels_first",
        input_shape=(3, 28, 28),
    )

    model = uft.generate_k_lip_model(
        SpectralConvTranspose2d, kwargs, kwargs["input_shape"], 1.0
    )

    # lay = SpectralConvTranspose2d(**kwargs)
    # model = Sequential([lay])
    x = np.random.normal(size=(5,) + kwargs["input_shape"])

    x = uft.to_tensor(x)
    y1 = model(x)

    # Test vanilla export inference comparison
    if uft.vanilla_require_a_copy():
        model2 = uft.generate_k_lip_model(
            SpectralConvTranspose2d, kwargs, kwargs["input_shape"], 1.0
        )
        uft.copy_model_parameters(model, model2)
        vanilla_model = uft.vanillaModel(model2)
    else:
        vanilla_model = uft.vanillaModel(model)  # .vanilla_export()
    y2 = vanilla_model(x)
    np.testing.assert_allclose(uft.to_numpy(y1), uft.to_numpy(y2), atol=1e-6)

    # Test saving/loading model
    with tempfile.TemporaryDirectory() as tmpdir:
        model_path = os.path.join(tmpdir, uft.MODEL_PATH)
        uft.save_model(model, model_path, overwrite=True)
        uft.load_model(
            model_path,
            layer_type=SpectralConvTranspose2d,
            layer_params=kwargs,
            input_shape=kwargs["input_shape"],
            k=1.0,
        )


@pytest.mark.parametrize(
    "layer_type,layer_params,k_lip_model,is_orthogonal,input_shape,lfactor,epochs",
    [
        (
            SpectralLinear,
            {"bias": False, "in_features": 4, "out_features": 3},
            1.0,
            True,
            (4,),
            1.0,
            5,
        ),
        (
            SpectralLinear,
            {"bias": False, "in_features": 4, "out_features": 3, "eps_bjorck": None},
            1.0,
            False,
            (4,),
            1.0,
            150,
        ),
        (
            FrobeniusLinear,
            {"bias": False, "in_features": 4, "out_features": 1},
            1.0,
            True,
            (4,),
            1.0,
            5,
        ),
        (
            FrobeniusLinear,
            {
                "bias": False,
                "in_features": 4,
                "out_features": 3,
                "disjoint_neurons": False,
            },
            1.0,
            False,
            (4,),
            1.0,
            5,
        ),
        (
            SpectralConv2d,
            {
                "in_channels": 3,
                "out_channels": 5,
                "kernel_size": (3, 3),
                "bias": False,
            },
            1.0,
            True,
            (3, 5, 5),
            3.0,  # conv factor
            5,
        ),
        (
            SpectralConv2d,
            {
                "in_channels": 3,
                "out_channels": 5,
                "kernel_size": (3, 3),
                "bias": False,
                "eps_bjorck": None,
            },
            1.0,
            False,
            (3, 5, 5),
            3.0,  # conv factor
            150,
        ),
        (
            FrobeniusConv2d,
            {
                "in_channels": 3,
                "out_channels": 5,
                "kernel_size": (3, 3),
                "bias": False,
            },
            1.0,
            False,
            (3, 5, 5),
            3.0,  # conv factor
            5,
        ),
        (
            SpectralConvTranspose2d,
            {
                "in_channels": 3,
                "out_channels": 5,
                "kernel_size": (3, 3),
                "bias": False,
            },
            1.0,
            True,
            (3, 5, 5),
            3.0,  # conv factor
            5,
        ),
        (
            SpectralConvTranspose2d,
            {
                "in_channels": 3,
                "out_channels": 5,
                "kernel_size": (3, 3),
                "bias": False,
                "eps_bjorck": None,
            },
            1.0,
            False,
            (3, 5, 5),
            3.0,  # conv factor
            150,
        ),
    ],
)
def test_ortho_kernel(
    layer_type, layer_params, k_lip_model, is_orthogonal, input_shape, lfactor, epochs
):
    test_params = dict(
        layer_type=layer_type,
        layer_params=layer_params,
        batch_size=25,
        steps_per_epoch=125,
        epochs=epochs,
        input_shape=input_shape,
        k_lip_data=1.0,
        k_lip_model=k_lip_model,
        callbacks=[],
    )
    model, kernel, loss_fn, metrics = build_and_train_k_lip_model(**test_params)
    weights = uft.get_layer_weights_by_index(model, 0)
    weights = uft.to_numpy(weights)
    weights = weights.reshape(weights.shape[0], -1)

    sigmas = np.linalg.svd(
        weights,
        full_matrices=False,
        compute_uv=False,
    )
    print(lfactor * sigmas)
    if is_orthogonal:
        np.testing.assert_allclose(lfactor * sigmas, np.ones_like(sigmas), atol=1e-5)
    else:
        assert lfactor * sigmas.max() < k_lip_model * 1.02, "sigma max is too large"
        diff_to_one = np.max(np.ones_like(sigmas) - lfactor * sigmas)
        assert diff_to_one > 0.1, f"Too close to orthogonality  {diff_to_one}"



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
import pytest
import os
import numpy as np
from . import utils_framework as uft
from .utils_framework import (
    CategoricalCrossentropy,
    GroupSort,
    HouseHolder,
)


def check_serialization(layer_type, layer_params):
    m = uft.generate_k_lip_model(layer_type, layer_params, input_shape=(10,), k=1)
    if m is None:
        return
    loss, optimizer, _ = uft.compile_model(
        m,
        optimizer=uft.get_instance_framework(uft.SGD, inst_params={"model": m}),
        loss=CategoricalCrossentropy(from_logits=True),
    )
    name = layer_type.__class__.__name__
    path = os.path.join("logs", "activations", name)
    xnp = np.random.uniform(-10, 10, (255, 10))
    x = uft.to_tensor(xnp)
    y1 = m(x)
    uft.save_model(m, path)
    m2 = uft.load_model(
        path,
        compile=True,
        layer_type=layer_type,
        layer_params=layer_params,
        input_shape=(10,),
        k=1,
    )
    y2 = m2(x)
    np.testing.assert_allclose(uft.to_numpy(y1), uft.to_numpy(y2))


def test_group_sort_simple():
    check_serialization(GroupSort, layer_params={"group_size": 2})
    check_serialization(GroupSort, layer_params={"group_size": 5})


@pytest.mark.parametrize(
    "group_size,img,expected",
    [
        (
            2,
            False,
            [
                [1.0, 2.0, 3.0, 4.0],
                [3.0, 4.0, 1.0, 2.0],
                [1.0, 2.0, 1.0, 2.0],
                [1.0, 2.0, 1.0, 2.0],
            ],
        ),
        (
            4,
            False,
            [
                [1.0, 2.0, 3.0, 4.0],
                [1.0, 2.0, 3.0, 4.0],
                [1.0, 1.0, 2.0, 2.0],
                [1.0, 1.0, 2.0, 2.0],
            ],
        ),
        (
            2,
            True,
            [
                [1.0, 2.0, 3.0, 4.0],
                [3.0, 4.0, 1.0, 2.0],
                [1.0, 2.0, 1.0, 2.0],
                [1.0, 2.0, 1.0, 2.0],
            ],
        ),
        (
            4,
            True,
            [
                [1.0, 2.0, 3.0, 4.0],
                [1.0, 2.0, 3.0, 4.0],
                [1.0, 1.0, 2.0, 2.0],
                [1.0, 1.0, 2.0, 2.0],
            ],
        ),
    ],
)
def test_GroupSort(group_size, img, expected):
    gs = uft.get_instance_framework(GroupSort, {"group_size": group_size})
    if gs is None:
        return
    x = [
        [1, 2, 3, 4],
        [4, 3, 2, 1],
        [1, 2, 1, 2],
        [2, 1, 2, 1],
    ]

    if not img:
        x = uft.to_tensor(x)
        uft.build_layer(gs, (4,))
    else:
        xn = np.asarray(x)
        xnp = np.repeat(
            np.expand_dims(np.repeat(np.expand_dims(xn, -1), 28, -1), -1), 28, -1
        )
        xnp = uft.to_NCHW_inv(xnp)  # move channel if needed (TF)
        x = uft.to_tensor(xnp)
        uft.build_layer(gs, (28, 28, 4))
    y = gs(x).numpy()
    y_t = expected
    if img:
        y_tnp = np.asarray(y_t)
        y_t = np.repeat(
            np.expand_dims(np.repeat(np.expand_dims(y_tnp, -1), 28, -1), -1), 28, -1
        )
        y_t = uft.to_NCHW_inv(y_t)  # move channel if needed (TF)
    np.testing.assert_equal(y, y_t)


@pytest.mark.parametrize("group_size", [2, 4])
def test_GroupSort_idempotence(group_size):
    gs = uft.get_instance_framework(GroupSort, {"group_size": group_size})
    if gs is None:
        return
    xnp = np.random.uniform(-10, 10, (255, 16))
    x = uft.to_tensor(xnp)
    uft.build_layer(gs, (16,))
    y1 = gs(x)
    y2 = gs(y1)
    np.testing.assert_equal(y1.numpy(), y2.numpy())


"""Tests for HouseHolder activation:
    - instantiation of layer
    - check outputs on dense (bs, n) tensor, with three thetas: 0, pi/2 and pi
    - check outputs on dense (bs, h, w, n) tensor, with three thetas: 0, pi/2 and pi
    - check idempotence hh(hh(x)) = hh(x)
"""


@pytest.mark.skipif(
    hasattr(HouseHolder, "unavailable_class"), reason="HouseHolder not available"
)
@pytest.mark.parametrize(
    "params,shape,len_shape,expected",
    [
        (
            {"channels": 10},
            (10, 28, 28),
            (5,),
            np.pi / 2,
        ),  # Instantiation without argument
        (
            {
                "data_format": "channels_last",
                "channels": 16,
                "k_coef_lip": 2.5,
                "theta_initializer": "ones",
            },
            (16, 32, 32),
            (8,),
            1,
        ),  # Instantiation with arguments
    ],
)
def test_HouseHolder_instantiation(params, shape, len_shape, expected):
    shape = uft.to_framework_channel(shape)
    hh = uft.get_instance_framework(HouseHolder, params)
    uft.build_layer(hh, shape)
    theta = np.squeeze(uft.to_numpy(hh.theta))
    assert theta.shape == len_shape
    np.testing.assert_allclose(theta, expected)


@pytest.mark.skipif(
    hasattr(HouseHolder, "unavailable_class"), reason="HouseHolder not available"
)
def test_HouseHolder_serialization():
    # Check serialization
    check_serialization(
        HouseHolder, layer_params={"channels": 10, "theta_initializer": "normal"}
    )

    if uft.framework == "torch":
        pytest.skip("data format skipped in  torch")
    # Instantiation error because of wrong data format
    with pytest.raises(RuntimeError):
        _ = uft.get_instance_framework(
            HouseHolder, {"channels": 4, "data_format": "channels_first"}
        )


@pytest.mark.skipif(
    hasattr(HouseHolder, "unavailable_class"), reason="HouseHolder not available"
)
@pytest.mark.parametrize("dense", [(True,), (False,)])
def test_HouseHolder_theta_zero(dense):
    """HouseHolder with theta=0 on 2-D tensor (bs, n).
    Theta=0 means Id if z2 > 0, and reflection if z2 < 0.
    """
    if dense:
        bs = np.random.randint(64, 512)
        n = np.random.randint(1, 1024) * 2
        size = (bs, n // 2)
        ch = n
    else:  # convolutional
        bs = np.random.randint(32, 128)
        h, w = np.random.randint(1, 64), np.random.randint(1, 64)
        c = np.random.randint(1, 64) * 2
        size = (bs,) + uft.to_framework_channel((c // 2, h, w))
        ch = c

    hh = uft.get_instance_framework(
        HouseHolder, {"channels": ch, "theta_initializer": "zeros"}
    )

    # Case 1: hh(x) = x   (identity case, z2 > 0)
    z1 = np.random.normal(size=size)
    z2 = np.random.uniform(size=size)
    x = np.concatenate([z1, z2], axis=-1)
    y = uft.to_numpy(hh(uft.to_tensor(x)))
    np.testing.assert_allclose(y, x)

    # Case 2: hh(x) = [z1, -z2]   (reflection across z1 axis, z2 < 0)
    z1 = np.random.normal(size=size)
    z2 = -np.random.uniform(size=size)
    x = np.concatenate([z1, z2], axis=-1)
    expected_output = np.concatenate([z1, -z2], axis=-1)
    y = uft.to_numpy(hh(uft.to_tensor(x)))
    np.testing.assert_allclose(y, expected_output)


@pytest.mark.skipif(
    hasattr(HouseHolder, "unavailable_class"), reason="HouseHolder not available"
)
@pytest.mark.parametrize("dense", [(True,), (False,)])
def test_HouseHolder_theta_pi(dense):
    """HouseHolder with theta=pi on 2-D tensor (bs, n).
    Theta=pi means Id if z1 < 0, and reflection if z1 > 0.
    """
    if dense:
        bs = np.random.randint(64, 512)
        n = np.random.randint(1, 1024) * 2
        size = (bs, n // 2)
        ch = n
    else:  # convolutional
        bs = np.random.randint(32, 128)
        h, w = np.random.randint(1, 64), np.random.randint(1, 64)
        c = np.random.randint(1, 64) * 2
        size = (bs,) + uft.to_framework_channel((c // 2, h, w))
        ch = c

    hh = uft.get_instance_framework(
        HouseHolder,
        {"channels": ch, "theta_initializer": uft.initializers_Constant(np.pi)},
    )
    # Case 1: hh(x) = x   (identity case, z1 < 0)
    z1 = -np.random.uniform(size=size)
    z2 = np.random.normal(size=size)
    x = np.concatenate([z1, z2], axis=-1)
    y = uft.to_numpy(hh(uft.to_tensor(x)))
    np.testing.assert_allclose(y, x, atol=1e-6)

    # Case 2: hh(x) = [z1, -z2]   (reflection across  z2 axis, z1 > 0)
    z1 = np.random.uniform(size=size)
    z2 = np.random.normal(size=size)
    x = np.concatenate([z1, z2], axis=-1)
    expected_output = np.concatenate([-z1, z2], axis=-1)
    y = uft.to_numpy(hh(uft.to_tensor(x)))
    np.testing.assert_allclose(y, expected_output, atol=1e-6)


@pytest.mark.skipif(
    hasattr(HouseHolder, "unavailable_class"), reason="HouseHolder not available"
)
@pytest.mark.parametrize("dense", [(True,), (False,)])
def test_HouseHolder_theta_90(dense):
    """HouseHolder with theta=pi/2 on 2-D tensor (bs, n).
    Theta=pi/2 is equivalent to GroupSort2: Id if z1 < z2, and reflection if z1 > z2
    """
    if dense:
        bs = np.random.randint(64, 512)
        n = np.random.randint(1, 1024) * 2
        size = (bs, n // 2)
        ch = n
    else:  # convolutional
        bs = np.random.randint(32, 128)
        h, w = np.random.randint(1, 64), np.random.randint(1, 64)
        c = np.random.randint(1, 64) * 2
        size = (bs,) + uft.to_framework_channel((c // 2, h, w))
        ch = c

    hh = uft.get_instance_framework(HouseHolder, {"channels": ch})
    # Case 1: hh(x) = x   (identity case, z1 < z2)
    z1 = -np.random.normal(size=size)
    z2 = z1 + np.random.uniform(size=size)
    x = np.concatenate([z1, z2], axis=-1)
    y = uft.to_numpy(hh(uft.to_tensor(x)))
    np.testing.assert_allclose(y, x)

    # Case 2: hh(x) = reflection(x)   (if z1 > z2)
    z1 = np.random.normal(size=size)
    z2 = z1 - np.random.uniform(size=size)
    x = np.concatenate([z1, z2], axis=-1)
    expected_output = np.concatenate([z2, z1], axis=-1)
    y = uft.to_numpy(hh(uft.to_tensor(x)))
    np.testing.assert_allclose(y, expected_output, atol=1e-6)


@pytest.mark.skipif(
    hasattr(HouseHolder, "unavailable_class"), reason="HouseHolder not available"
)
def test_HouseHolder_idempotence():
    """Assert idempotence of HouseHolder activation: hh(hh(x)) = hh(x)"""

    bs = np.random.randint(32, 128)
    h, w = np.random.randint(1, 64), np.random.randint(1, 64)
    c = np.random.randint(1, 32) * 2
    hh = uft.get_instance_framework(
        HouseHolder, {"channels": c, "theta_initializer": "normal"}
    )
    x = np.random.normal(size=(bs,) + uft.to_framework_channel((c, h, w)))
    x = uft.to_tensor(x)

    # Run two times the HH activation and compare both outputs
    y = hh(x)
    z = hh(y)
    np.testing.assert_allclose(uft.to_numpy(y), uft.to_numpy(z))



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
from typing import Tuple, Union

import numpy as np
import torch
import torch.nn as nn
import torch.nn.utils.parametrize as parametrize


def compute_lconv_coef_1d(
    kernel_size: Tuple[int],
    input_shape: Tuple[int] = None,
    strides: Tuple[int] = (1,),
    padding_mode: str = "zeros",
) -> float:
    stride = strides[0]
    k1 = kernel_size[0]

    if (padding_mode in ["zeros"]) and (stride == 1) and (input_shape is not None):
        # See https://arxiv.org/abs/2006.06520
        in_l = input_shape[-1]
        k1_div2 = (k1 - 1) / 2
        coefLip = in_l / (k1 * in_l - k1_div2 * (k1_div2 + 1))
    else:
        sn1 = strides[0]
        coefLip = 1.0 / np.ceil(k1 / sn1)

    return coefLip  # type: ignore


def compute_lconv_coef(
    kernel_size: Tuple[int, ...],
    input_shape: Tuple[int, ...] = None,
    strides: Tuple[int, ...] = (1, 1),
    padding_mode: str = "zeros",
) -> float:
    # See https://arxiv.org/abs/2006.06520
    stride = np.prod(strides)
    k1, k2 = kernel_size

    if (padding_mode in ["zeros"]) and (stride == 1) and (input_shape is not None):
        h, w = input_shape[-2:]
        k1_div2 = (k1 - 1) / 2
        k2_div2 = (k2 - 1) / 2
        coefLip = np.sqrt(
            (w * h)
            / ((k1 * h - k1_div2 * (k1_div2 + 1)) * (k2 * w - k2_div2 * (k2_div2 + 1)))
        )
    else:
        sn1 = strides[0]
        sn2 = strides[1]
        coefLip = np.sqrt(1.0 / (np.ceil(k1 / sn1) * np.ceil(k2 / sn2)))

    return coefLip  # type: ignore


class _LConvNorm(nn.Module):
    """Parametrization module for kernel normalization of lipschitz convolution."""

    def __init__(self, lconv_coefficient: float) -> None:
        super().__init__()
        self.lconv_coefficient = lconv_coefficient

    def forward(self, weight: torch.Tensor) -> torch.Tensor:
        return weight * self.lconv_coefficient


ConvType = Union[torch.nn.Conv2d, torch.nn.Conv1d]


def lconv_norm(module: ConvType, name: str = "weight") -> ConvType:
    r"""
    Applies Lipschitz normalization to a kernel in the given convolutional.
    This is implemented via a hook that multiplies the kernel by a value computed
    from the input shape before every ``forward()`` call.

    See `Achieving robustness in classification using optimal transport with hinge
    regularization <https://arxiv.org/abs/2006.06520>`_.

    Args:
        module: Containing module.
        name: Name of weight parameter.

    Returns:
        The original module with the Lipschitz normalization hook.

    Example::

        >>> m = lconv_norm(nn.Conv2d(16, 16, (3, 3)))
        >>> m
        Conv2d(16, 16, kernel_size=(3, 3), stride=(1, 1))

    """
    onedim = isinstance(module, torch.nn.Conv1d)
    if onedim:
        coefficient = compute_lconv_coef_1d(module.kernel_size, None, module.stride)
    else:
        coefficient = compute_lconv_coef(module.kernel_size, None, module.stride)
    parametrize.register_parametrization(module, name, _LConvNorm(coefficient))
    return module


def remove_lconv_norm(module: torch.nn.Conv2d, name: str = "weight") -> torch.nn.Conv2d:
    r"""
    Removes the normalization parametrization for lipschitz convolution from a module.

    Args:
        module: Containing module.
        name: Name of weight parameter.

    Example:

        >>> m = lconv_norm(nn.Conv2d(16, 16, (3, 3)))
        >>> remove_lconv_norm(m)
    """
    for key, m in module.parametrizations[name]._modules.items():
        if isinstance(m, _LConvNorm):
            if len(module.parametrizations[name]) == 1:
                parametrize.remove_parametrizations(module, name)
            else:
                del module.parametrizations[name]._modules[key]



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""Tests for singular value computation (in compute_layer_sv.py)"""

import os
import pprint
import pytest

import numpy as np

from .utils_framework import compute_layer_sv
from .utils_framework import (
    tLinear,
    FrobeniusConv2d,
    FrobeniusLinear,
    SpectralConv2d,
    SpectralLinear,
)
from .utils_framework import OrthLinearRegularizer

from . import utils_framework as uft

pp = pprint.PrettyPrinter(indent=4)


def linear_generator(batch_size, input_shape: tuple, kernel):
    """
    Generate data according to a linear kernel
    Args:
        batch_size: size of each batch
        input_shape: shape of the desired input
        kernel: kernel used to generate data, must match the last dimensions of
            `input_shape`

    Returns:
        a generator for the data

    """
    input_shape = tuple(input_shape)
    while True:
        # pick random sample in [0, 1] with the input shape
        batch_x = np.array(
            np.random.uniform(-10, 10, (batch_size,) + input_shape), dtype=np.float16
        )
        # apply the k lip linear transformation
        batch_y = np.tensordot(
            batch_x,
            kernel,
            axes=(
                [i for i in range(1, len(input_shape) + 1)],
                [i for i in range(0, len(input_shape))],
            ),
        )
        yield batch_x, batch_y


def build_kernel(input_shape: tuple, output_shape: tuple, k=1.0):
    """
    build a kernel with defined lipschitz factor

    Args:
        input_shape: input shape of the linear function
        output_shape: output shape of the linear function
        k: lipshitz factor of the function

    Returns:
        the kernel for use in the linear_generator

    """
    input_shape = tuple(input_shape)
    output_shape = tuple(output_shape)
    kernel = np.array(
        np.random.random_sample(input_shape + output_shape), dtype=np.float16
    )
    kernel = (
        kernel * k / np.linalg.norm(kernel)
    )  # assuming lipschitz constraint is independent with respect to the chosen metric

    return kernel


def train_compute_and_verifySV(
    layer_type: type,
    layer_params: dict,
    batch_size: int,
    steps_per_epoch: int,
    epochs: int,
    input_shape: tuple,
    k_lip_model: float,
    k_lip_data: float,
    **kwargs
):
    """
    Create a  model, train compute and verify SVs.

    Args:
        layer_type:
        layer_params:
        batch_size:
        steps_per_epoch:
        epochs:
        input_shape:
        k_lip_model:
        k_lip_data:
        **kwargs:

    Returns:
    """
    flag_test_SVmin = True
    if "test_SVmin" in kwargs.keys():
        flag_test_SVmin = kwargs["test_SVmin"]
    if "k_lip_tolerance_factor" not in kwargs.keys():
        kwargs["k_lip_tolerance_factor"] = 1.02
    # clear session to avoid side effects from previous train
    uft.init_session()  # K.clear_session()
    np.random.seed(42)
    input_shape = uft.to_framework_channel(input_shape)

    # tf.random.uft.set_seed(1234)
    # create the model, defin opt, and compile it
    model = uft.generate_k_lip_model(layer_type, layer_params, input_shape, k_lip_model)

    optimizer = uft.get_instance_framework(
        uft.Adam, inst_params={"lr": 0.001, "model": model}
    )
    # optimizer = uft.Adam(lr=0.001)
    loss_fn, optimizer, metrics = uft.compile_model(
        model,
        optimizer=optimizer,
        loss=uft.MeanSquaredError(),
        metrics=[uft.metric_mse()],
    )
    # model.compile(
    #     optimizer=optimizer, loss="mean_squared_error", metrics=[metrics.mse]
    # )
    # create the synthetic data generator
    output_shape = uft.compute_output_shape(input_shape, model)
    # output_shape = model.uft.compute_output_shape((batch_size,) + input_shape)[1:]
    kernel = build_kernel(input_shape, output_shape, k_lip_data)
    # define logging features
    logdir = os.path.join("logs", uft.LIP_LAYERS, "%s" % layer_type.__name__)
    os.makedirs(logdir, exist_ok=True)

    callback_list = []
    if "callbacks" in kwargs and (kwargs["callbacks"] is not None):
        callback_list = callback_list + kwargs["callbacks"]
    # train model

    traind_ds = linear_generator(batch_size, input_shape, kernel)
    uft.train(
        traind_ds,
        model,
        loss_fn,
        optimizer,
        epochs,
        batch_size,
        steps_per_epoch=steps_per_epoch,
        callbacks=callback_list,
    )
    # model.fit(
    #     linear_generator(batch_size, input_shape, kernel),
    #     steps_per_epoch=steps_per_epoch,
    #     epochs=epochs,
    #     verbose=1,
    #     callbacks=callback_list,
    # )

    for ll in uft.get_children(model):  # .layers:
        SVmin, SVmax = compute_layer_sv(ll)
        # log metrics
        if SVmin is not None:
            assert SVmax < k_lip_model * kwargs["k_lip_tolerance_factor"], (
                " the maximum singular value of the layer "
                + ll.name
                + " must be lower than the specified boundary"
            )  # noqa: E501

            assert SVmin <= SVmax, (
                " the minimum singular value of the layer "
                + ll.name
                + " must be lower than the maximum value"
            )  # noqa: E501
            if flag_test_SVmin:
                assert SVmin > k_lip_model * (2.0 - kwargs["k_lip_tolerance_factor"]), (
                    " the minimum singular value of the layer "
                    + ll.name
                    + " must be greater than the specified boundary"
                )  # noqa: E501
    return


@pytest.mark.skipif(
    hasattr(compute_layer_sv, "unavailable_class"),
    reason="compute_layer_sv not available",
)
@pytest.mark.parametrize(
    "test_params",
    [
        dict(
            layer_type=SpectralLinear,
            layer_params={"bias": False, "in_features": 4, "out_features": 4},
            batch_size=1000,
            steps_per_epoch=125,
            epochs=5,
            input_shape=(4,),
            k_lip_data=1.0,
            k_lip_model=1.0,
            callbacks=[],
        ),
        dict(
            layer_type=SpectralLinear,
            layer_params={"in_features": 4, "out_features": 4},
            batch_size=1000,
            steps_per_epoch=125,
            epochs=5,
            input_shape=(4,),
            k_lip_data=1.0,
            k_lip_model=5.0,
            callbacks=[],
        ),
        dict(
            layer_type=FrobeniusLinear,
            layer_params={"in_features": 4, "out_features": 1},
            batch_size=1000,
            steps_per_epoch=125,
            epochs=5,
            input_shape=(4,),
            k_lip_data=1.0,
            k_lip_model=1.0,
            test_SVmin=False,
            callbacks=[],
        ),
        dict(
            layer_type=FrobeniusLinear,
            layer_params={"in_features": 4, "out_features": 1},
            batch_size=1000,
            steps_per_epoch=125,
            epochs=5,
            input_shape=(4,),
            k_lip_data=1.0,
            k_lip_model=5.0,
            test_SVmin=False,
            callbacks=[],
        ),
        dict(
            layer_type=SpectralConv2d,
            layer_params={
                "in_channels": 1,
                "out_channels": 2,
                "kernel_size": (3, 3),
                "bias": False,
            },
            batch_size=100,
            steps_per_epoch=125,
            epochs=5,
            input_shape=(1, 5, 5),
            k_lip_data=1.0,
            k_lip_model=1.0,
            k_lip_tolerance_factor=1.02,
            test_SVmin=False,
            callbacks=[],
        ),
        dict(
            layer_type=SpectralConv2d,
            layer_params={
                "in_channels": 1,
                "out_channels": 2,
                "kernel_size": (3, 3),
            },
            batch_size=100,
            steps_per_epoch=125,
            epochs=5,
            input_shape=(1, 5, 5),
            k_lip_data=1.0,
            k_lip_model=5.0,
            k_lip_tolerance_factor=1.02,
            test_SVmin=False,
            callbacks=[],
        ),
        dict(
            layer_type=SpectralConv2d,
            layer_params={
                "in_channels": 3,
                "out_channels": 2,
                "kernel_size": (3, 3),
                "bias": False,
            },
            batch_size=100,
            steps_per_epoch=125,
            epochs=5,
            input_shape=(3, 5, 5),  # case conv_first=False
            k_lip_data=1.0,
            k_lip_model=1.0,
            k_lip_tolerance_factor=1.02,
            test_SVmin=False,
            callbacks=[],
        ),
        dict(
            layer_type=SpectralConv2d,
            layer_params={
                "in_channels": 1,
                "out_channels": 5,
                "kernel_size": (3, 3),
                "bias": False,
                "stride": 2,
            },
            batch_size=100,
            steps_per_epoch=125,
            epochs=5,
            input_shape=(1, 10, 10),
            k_lip_data=1.0,
            k_lip_model=1.0,
            k_lip_tolerance_factor=1.02,
            test_SVmin=False,
            callbacks=[],
        ),
        dict(
            layer_type=SpectralConv2d,
            layer_params={
                "in_channels": 1,
                "out_channels": 3,
                "kernel_size": (3, 3),
                "bias": False,
                "stride": 2,
            },
            batch_size=100,
            steps_per_epoch=125,
            epochs=5,
            input_shape=(1, 10, 10),
            k_lip_data=1.0,
            k_lip_model=1.0,
            k_lip_tolerance_factor=1.02,
            test_SVmin=False,
            callbacks=[],
        ),
        dict(
            layer_type=FrobeniusConv2d,
            layer_params={
                "in_channels": 1,
                "out_channels": 2,
                "kernel_size": (3, 3),
            },
            batch_size=100,
            steps_per_epoch=125,
            epochs=5,
            input_shape=(1, 5, 5),
            k_lip_data=1.0,
            k_lip_model=1.0,
            k_lip_tolerance_factor=1.1,  # Frobenius seems less precise on SVs
            test_SVmin=False,
            callbacks=[],
        ),
        dict(
            layer_type=FrobeniusConv2d,
            layer_params={
                "in_channels": 1,
                "out_channels": 2,
                "kernel_size": (3, 3),
                "bias": False,
            },
            batch_size=100,
            steps_per_epoch=125,
            epochs=5,
            input_shape=(1, 5, 5),
            k_lip_data=1.0,
            k_lip_model=5.0,
            k_lip_tolerance_factor=1.1,
            test_SVmin=False,
            callbacks=[],
        ),
        dict(
            layer_type=FrobeniusConv2d,
            layer_params={
                "in_channels": 3,
                "out_channels": 2,
                "kernel_size": (3, 3),
            },
            batch_size=100,
            steps_per_epoch=125,
            epochs=5,
            input_shape=(3, 5, 5),  # case conv_first=False
            k_lip_data=1.0,
            k_lip_model=1.0,
            k_lip_tolerance_factor=1.1,  # Frobenius seems less precise on SVs
            test_SVmin=False,
            callbacks=[],
        ),
    ],
)
def test_verifySV(test_params):
    train_compute_and_verifySV(**test_params)


@pytest.mark.skipif(
    hasattr(compute_layer_sv, "unavailable_class")
    or hasattr(OrthLinearRegularizer, "unavailable_class"),
    reason="compute_layer_sv or OrthLinearRegularizer not available",
)
@pytest.mark.parametrize(
    "test_params",
    [
        dict(
            layer_type=tLinear,
            layer_params={
                "in_features": 4,
                "out_features": 6,
                "kernel_regularizer": uft.get_instance_framework(
                    OrthLinearRegularizer, {"lambda_orth": 1000.0}
                ),
            },
            batch_size=1000,
            steps_per_epoch=125,
            epochs=10,
            input_shape=(4,),
            k_lip_data=1.0,
            k_lip_model=1.0,
            callbacks=[],
        ),
    ],
)
def test_verifySV_orthRegul(test_params):
    train_compute_and_verifySV(**test_params)



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
import pytest
import numpy as np
import os

from . import utils_framework as uft

from .utils_framework import (
    tLinear,
    CategoricalProvableRobustAccuracy,
    BinaryProvableRobustAccuracy,
    CategoricalProvableAvgRobustness,
    BinaryProvableAvgRobustness,
)


def one_hot_data(x, n_class):
    """Return a Framework float32 tensor of shape [N, n_class]
    from a list/np.array of shape [N]"""
    return np.eye(n_class)[x]


@pytest.mark.parametrize(
    "nb_class, loss, loss_params, nb_classes",
    [
        (
            10,
            CategoricalProvableRobustAccuracy,
            {"epsilon": 1, "lip_const": 1.0, "disjoint_neurons": False},
            10,
        ),
        (
            10,
            BinaryProvableRobustAccuracy,
            {
                "epsilon": 1,
                "lip_const": 1.0,
            },
            10,
        ),
        (
            10,
            CategoricalProvableAvgRobustness,
            {"lip_const": 2.0, "disjoint_neurons": False, "negative_robustness": True},
            10,
        ),
        (
            10,
            BinaryProvableAvgRobustness,
            {"lip_const": 2.0, "negative_robustness": True},
            10,
        ),
    ],
)
def test_serialization(nb_class, loss, loss_params, nb_classes):
    if hasattr(loss, "unavailable_class"):
        pytest.skip(f"{loss} not implemented")

    n = 255
    x = np.random.uniform(size=(n, 42))
    y = one_hot_data(np.random.randint(nb_classes, size=n), nb_classes)
    x, y = uft.to_tensor(x), uft.to_tensor(y)
    m = uft.generate_k_lip_model(
        tLinear,
        layer_params={"in_features": 42, "out_features": nb_class},
        input_shape=(42,),
    )

    loss_fn, optimizer, metrics = uft.compile_model(
        m,
        optimizer=uft.get_instance_framework(
            uft.SGD, inst_params={"lr": 0.001, "model": m}
        ),
        loss=uft.get_instance_framework(loss, inst_params=loss_params),
    )

    l1 = m.evaluate(x, y)
    name = loss.__class__.__name__
    path = os.path.join("logs", "metrics", name)

    uft.save_model(m, path)
    m2 = uft.load_model(
        path,
        compile=True,
        layer_type=tLinear,
        layer_params={"in_features": 42, "out_features": nb_class},
        input_shape=(42,),
        k=1,
    )
    l2 = m2.evaluate(x, y)
    np.testing.assert_equal(
        l1,
        l2,
        err_msg=f"serialization changed loss value for {loss}",
    )
    return


def build_test_provable_vs_adjusted():
    test_data = []
    if not hasattr(CategoricalProvableAvgRobustness, "unavailable_class"):
        for lip_cst in [1, 0.5, 2.0]:
            for disjoint in [True, False]:
                test_data.append(
                    (
                        CategoricalProvableAvgRobustness,
                        {
                            "lip_const": lip_cst,
                            "disjoint_neurons": disjoint,
                            "negative_robustness": True,
                        },
                        10,
                    )
                )

    if not hasattr(BinaryProvableAvgRobustness, "unavailable_class"):
        for lip_cst in [1.0, 0.5, 2.0]:
            test_data.append(
                (
                    BinaryProvableAvgRobustness,
                    {"lip_const": lip_cst, "negative_robustness": True},
                    1,
                )
            )
    return test_data


@pytest.mark.parametrize(
    "loss, loss_params, nb_class", build_test_provable_vs_adjusted()
)
def test_provable_vs_adjusted(loss, loss_params, nb_class):
    n = 500
    x = np.random.uniform(size=(n, nb_class))
    if nb_class > 1:
        y = one_hot_data(np.random.randint(nb_class, size=n), nb_class)
    else:
        y = np.random.randint(2, size=n)
    x, y = uft.to_tensor(x), uft.to_tensor(y)

    pr = uft.get_instance_framework(loss, inst_params=loss_params)
    if pr is None:
        pytest.skip(f"{loss} not implemented")
    other_param = loss_params.copy()
    other_param["negative_robustness"] = not loss_params["negative_robustness"]
    ar = uft.get_instance_framework(loss, inst_params=other_param)
    l1 = pr(y, y).numpy()
    l2 = ar(y, y).numpy()
    np.testing.assert_allclose(
        l1,
        l2,
        1e-4,
        err_msg=f"{loss} provable and adjusted "
        "robustness must give same "
        "values when y_true==y_pred",
    )

    l1 = pr(y, x).numpy()
    l2 = ar(y, x).numpy()
    diff = np.min(np.abs(l1 - l2))
    assert diff > 1e-4, f"{loss} provable and adjusted robustness must give \
        different values when y_true!=y_pred"


@pytest.mark.parametrize(
    "loss, loss_params, nb_class",
    [
        (
            CategoricalProvableAvgRobustness,
            {
                "lip_const": 1.0,
                "disjoint_neurons": True,
                "negative_robustness": True,
            },
            10,
        ),
        (
            BinaryProvableAvgRobustness,
            {"lip_const": 1.0, "negative_robustness": True},
            1,
        ),
    ],
)
def test_data_format(loss, loss_params, nb_class):
    n = 500

    if hasattr(loss, "unavailable_class"):
        pytest.skip(f"{loss} not implemented")
    pr = uft.get_instance_framework(loss, inst_params=loss_params)

    other_param = loss_params.copy()
    other_param["negative_robustness"] = not loss_params["negative_robustness"]
    ar = uft.get_instance_framework(loss, inst_params=other_param)

    for metric in [pr, ar]:
        x = np.random.uniform(size=(n, nb_class))
        if nb_class > 1:
            y = one_hot_data(np.random.randint(nb_class, size=n), nb_class)
        else:
            y = np.random.randint(2, size=n)
        metrics_values = []
        for neg_val in [0.0, -1.0]:
            y_pred_case = x
            y_true_case = np.where(y > 0, 1.0, neg_val)
            y_pred_case, y_true_case = (
                uft.to_tensor(y_pred_case),
                uft.to_tensor(y_true_case),
            )
            metrics_values.append(metric(y_true_case, y_pred_case).numpy())
        assert all(
            [m == metrics_values[0] for m in metrics_values]
        ), f"{loss} changing the data format must not change the metric value"


@pytest.mark.parametrize(
    "loss, loss_params, nb_class",
    [
        (
            CategoricalProvableAvgRobustness,
            {
                "lip_const": 1.0,
                "disjoint_neurons": False,
                "negative_robustness": True,
            },
            10,
        ),
        (
            CategoricalProvableAvgRobustness,
            {
                "lip_const": 1.0,
                "disjoint_neurons": False,
                "negative_robustness": False,
            },
            10,
        ),
    ],
)
def test_disjoint_neurons(loss, loss_params, nb_class):
    n = 500

    if hasattr(loss, "unavailable_class"):
        pytest.skip(f"{loss} not implemented")

    pr = uft.get_instance_framework(loss, inst_params=loss_params)
    # if pr is None:
    #     pytest.skip(f"{loss} not implemented")
    other_param = loss_params.copy()
    other_param["disjoint_neurons"] = True
    pdr = uft.get_instance_framework(loss, inst_params=other_param)

    x = np.random.uniform(size=(n, nb_class))
    if nb_class > 1:
        y = one_hot_data(np.random.randint(nb_class, size=n), nb_class)
    else:
        y = np.random.randint(2, size=n)
    metrics_values = []
    for neg_val in [0.0, -1.0]:
        y_pred_case = x
        y_true_case = np.where(y > 0, 1.0, neg_val)
        y_pred_case, y_true_case = (
            uft.to_tensor(y_pred_case),
            uft.to_tensor(y_true_case),
        )
        metrics_values.append(pr(y_true_case, y_pred_case).numpy())
        metrics_values.append(
            pdr(y_true_case, y_pred_case * (np.sqrt(2) / 2.0)).numpy()
        )
    np.testing.assert_allclose(
        metrics_values[0],
        metrics_values,
        4,
        err_msg=f"{loss} does not account disjoint_neuron properly",
    )


y_pred1 = [
    [1.0, 0.0, 0.0],  # good class & over 0.25
    [0.1, 0.0, 0.0],  # good class & below 0.25
    [0.0, 0.0, 1.0],  # wrong class & over 0.25
    [0.0, 0.0, 0.1],  # wrong class & below 0.25
]
y_true1 = [
    [1.0, 0.0, 0.0],
    [1.0, 0.0, 0.0],
    [0.0, 1.0, 0.0],
    [0.0, 1.0, 0.0],
]
y_pred2 = [
    1.0,  # good class & over 0.25
    0.1,  # good class & below 0.25
    1.0,  # wrong class & over 0.25
    0.1,  # wrong class & below 0.25
    -1.0,
    -0.1,
    -1.0,
    -0.1,
]
y_true2 = [1.0, 1.0, -1.0, -1.0, -1.0, -1.0, 1.0, 1.0]


@pytest.mark.parametrize(
    "loss, loss_params, y_pred, y_true, expected_value",
    [
        (
            CategoricalProvableRobustAccuracy,
            {
                "epsilon": 0.25,
                "lip_const": 1.0,
                "disjoint_neurons": False,
            },
            y_pred1,
            y_true1,
            0.25,
        ),
        (
            CategoricalProvableAvgRobustness,
            {
                "lip_const": 1.0,
                "disjoint_neurons": False,
            },
            y_pred1,
            y_true1,
            0.25 * 1.1 / np.sqrt(2),
        ),
        (
            BinaryProvableRobustAccuracy,
            {
                "epsilon": 0.25,
                "lip_const": 1.0,
            },
            y_pred2,
            y_true2,
            0.25,
        ),
        (
            BinaryProvableAvgRobustness,
            {
                "lip_const": 1.0,
                "negative_robustness": False,
            },
            y_pred2,
            y_true2,
            0.125 * (1.1 * 2),
        ),
    ],
)
def test_hardcoded_values(loss, loss_params, y_pred, y_true, expected_value):
    if hasattr(loss, "unavailable_class"):
        pytest.skip(f"{loss} not implemented")
    pr = uft.get_instance_framework(loss, inst_params=loss_params)
    y_pred, y_true = uft.to_tensor(y_pred), uft.to_tensor(y_true)
    val = pr(y_true, y_pred).numpy()
    np.testing.assert_allclose(
        val,
        expected_value,
        5,
        err_msg=f"{loss}  did not get the expected result with hardcoded values ",
    )



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
from os import path

import setuptools

# read the contents of your README file

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

with open(path.join(this_directory, "deel/torchlip/VERSION")) as f:
    version = f.read().strip()

dev_requires = ["tox", "black", "flake8", "flake8-black", "numpy", "torch_testing"]

docs_requires = [
    "sphinx==3.3.1",
    "recommonmark",
    "sphinx_rtd_theme",
    "sphinx_markdown_builder",
    "ipython",  # required for Pygments
    "nbsphinx",
    "sphinxcontrib_katex",
    "pytorch_sphinx_theme",
]

setuptools.setup(
    name="deel-torchlip",
    version=version,
    author=", ".join(
        [
            "Mathieu SERRURIER",
            "Franck MAMALET",
            "Thibaut BOISSIN",
            "Mikaël CAPELLE",
            "Justin PLAKOO",
        ]
    ),
    author_email=", ".join(
        [
            "mathieu.serrurier@irt-saintexupery.com",
            "franck.mamalet@irt-saintexupery.com",
            "thibaut.boissin@irt-saintexupery.com",
            "justin.plakoo@irt-saintexupery.com",
        ]
    ),
    description="PyTorch implementation for k-Lipschitz layers",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/deel-ai/deel-torchlip",
    packages=setuptools.find_namespace_packages(include=["deel.*"]),
    include_package_data=True,
    install_requires=[
        "numpy",
        "torch",
    ],
    license="MIT",
    extras_require={"dev": dev_requires, "doc": docs_requires},
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
import math
from typing import Optional

import numpy as np
import torch
import torch.nn.functional as F
from torch.nn.common_types import _size_2_t

from .module import LipschitzModule


def computePoolScalingFactor(kernel_size):
    if isinstance(kernel_size, tuple):
        scalingFactor = math.sqrt(np.prod(np.asarray(kernel_size)))
    else:
        scalingFactor = kernel_size
    return scalingFactor


class ScaledAvgPool2d(torch.nn.AvgPool2d, LipschitzModule):
    def __init__(
        self,
        kernel_size: _size_2_t,
        stride: Optional[_size_2_t] = None,
        padding: _size_2_t = 0,
        ceil_mode: bool = False,
        count_include_pad: bool = True,
        divisor_override: bool = None,
        k_coef_lip: float = 1.0,
    ):
        """
        Average pooling operation for spatial data, but with a lipschitz bound.

        Args:
            kernel_size: The size of the window.
            stride: The stride of the window. Must be None or equal to
                ``kernel_size``. Default value is ``kernel_size``.
            padding: Implicit zero-padding to be added on both sides. Must
                be zero.
            ceil_mode: When True, will use ceil instead of floor to compute the output
                shape.
            count_include_pad: When True, will include the zero-padding in the averaging
                calculation.
            divisor_override: If specified, it will be used as divisor, otherwise
                ``kernel_size`` will be used.
            k_coef_lip: The Lipschitz factor to ensure. The output will be scaled
                by this factor.

        This documentation reuse the body of the original torch.nn.AveragePooling2D
        doc.
        """
        torch.nn.AvgPool2d.__init__(
            self,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            ceil_mode=ceil_mode,
            count_include_pad=count_include_pad,
            divisor_override=divisor_override,
        )
        LipschitzModule.__init__(self, k_coef_lip)

        self.scalingFactor = computePoolScalingFactor(self.kernel_size)

        if self.stride != self.kernel_size:
            raise RuntimeError("stride must be equal to kernel_size.")
        if np.sum(self.padding) != 0:
            raise RuntimeError(f"{type(self)} does not support padding.")

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        coeff = self._coefficient_lip * self.scalingFactor
        return torch.nn.AvgPool2d.forward(self, input) * coeff

    def vanilla_export(self):
        return self


class ScaledAdaptiveAvgPool2d(torch.nn.AdaptiveAvgPool2d, LipschitzModule):
    def __init__(
        self,
        output_size: _size_2_t,
        k_coef_lip: float = 1.0,
    ):
        """
        Applies a 2D adaptive average pooling over an input signal composed of several
        input planes.

        The output is of size H x W, for any input size.
        The number of output features is equal to the number of input planes.

        Args:
            output_size: The target output size of the image of the form H x W.
                Can be a tuple (H, W) or a single H for a square image H x H.
                H and W can be either a ``int``, or ``None`` which means the
                size will be the same as that of the input.
            k_coef_lip: The Lipschitz factor to ensure. The output will be scaled
                by this factor.

        This documentation reuse the body of the original
        nn.AdaptiveAvgPool2d doc.
        """
        if not isinstance(output_size, tuple) or len(output_size) != 2:
            raise RuntimeError("output_size must be a tuple of 2 integers")
        else:
            if output_size[0] != 1 or output_size[1] != 1:
                raise RuntimeError("output_size must be (1, 1) for Lipschitz constant")
        torch.nn.AdaptiveAvgPool2d.__init__(self, output_size)
        LipschitzModule.__init__(self, k_coef_lip)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        coeff = math.sqrt(input.shape[-2] * input.shape[-1]) * self._coefficient_lip
        return torch.nn.AdaptiveAvgPool2d.forward(self, input) * coeff

    def vanilla_export(self):
        return self


class ScaledL2NormPool2d(torch.nn.LPPool2d, LipschitzModule):
    def __init__(
        self,
        kernel_size: _size_2_t,
        stride: Optional[_size_2_t] = None,
        ceil_mode: bool = False,
        k_coef_lip: float = 1.0,
    ):
        """
        Average pooling operation for spatial data, with a lipschitz bound. This
        pooling operation is norm preserving (gradient=1 almost everywhere).

        [1] Y.-L.Boureau, J.Ponce, et Y.LeCun, « A Theoretical Analysis of Feature
        Pooling in Visual Recognition »,p.8.

        Args:
            kernel_size: The size of the window.
            stride: The stride of the window. Must be None or equal to
                ``kernel_size``. Default value is ``kernel_size``.
            ceil_mode: When True, will use ceil instead of floor to compute the output
                shape.
            k_coef_lip: The lipschitz factor to ensure. The output will be
                scaled by this factor.
        """
        torch.nn.LPPool2d.__init__(
            self,
            2,  # Norm 2
            kernel_size=kernel_size,
            stride=stride,
            ceil_mode=ceil_mode,
        )
        LipschitzModule.__init__(self, k_coef_lip)

        if (self.stride is not None) and (self.stride != self.kernel_size):
            raise RuntimeError("stride must be equal to kernel_size.")

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        coeff = self._coefficient_lip
        return torch.nn.LPPool2d.forward(self, input) * coeff

    def vanilla_export(self):
        if self._coefficient_lip == 1.0:
            return torch.nn.LPPool2d(
                2,  # Norm 2
                kernel_size=self.kernel_size,
                stride=self.stride,
                ceil_mode=self.ceil_mode,
            )
        else:
            return self


class ScaledAdaptativeL2NormPool2d(
    torch.nn.modules.pooling._AdaptiveAvgPoolNd, LipschitzModule
):
    def __init__(
        self,
        output_size: _size_2_t = (1, 1),
        k_coef_lip: float = 1.0,
    ):
        """
        Average pooling operation for spatial data, with a lipschitz bound. This
        pooling operation is norm preserving (aka gradient=1 almost everywhere).

        [1]Y.-L.Boureau, J.Ponce, et Y.LeCun, « A Theoretical Analysis of Feature
        Pooling in Visual Recognition »,p.8.

        Arguments:
            output_size: the target output size has to be (1,1)
            k_coef_lip: the lipschitz factor to ensure

        Input shape:
            4D tensor with shape `(batch_size, channels, rows, cols)`.

        Output shape:
            4D tensor with shape `(batch_size, channels, 1, 1)`.
        """
        if not isinstance(output_size, tuple) or len(output_size) != 2:
            raise RuntimeError("output_size must be a tuple of 2 integers")
        else:
            if output_size[0] != 1 or output_size[1] != 1:
                raise RuntimeError("output_size must be (1, 1)")
        torch.nn.modules.pooling._AdaptiveAvgPoolNd.__init__(self, output_size)
        LipschitzModule.__init__(self, k_coef_lip)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return F.lp_pool2d(input, 2, input.shape[-2:]) * self._coefficient_lip

    def vanilla_export(self):
        return self



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
# flake8: noqa

from os import path

with open(path.join(path.dirname(__file__), "VERSION")) as f:
    __version__ = f.read().strip()

from . import functional, init, normalizers, utils
from .modules import *

__all__ = [
    "FrobeniusConv2d",
    "FrobeniusLinear",
    "FullSort",
    "GroupSort",
    "GroupSort2",
    "HKRLoss",
    "HKRMulticlassLoss",
    "SoftHKRMulticlassLoss",
    "LseHKRMulticlassLoss",
    "HingeMarginLoss",
    "HingeMulticlassLoss",
    "InvertibleDownSampling",
    "InvertibleUpSampling",
    "KRLoss",
    "KRMulticlassLoss",
    "LipschitzModule",
    "LipschitzPReLU",
    "MaxMin",
    "NegKRLoss",
    "ScaledAdaptiveAvgPool2d",
    "ScaledAvgPool2d",
    "ScaledL2NormPooling2D",
    "Sequential",
    "SpectralConv2d",
    "SpectralLinear",
    "vanilla_model",
]
